import os

from glob import glob
from bidict import OrderedBidict
from simplejson import OrderedDict
import ruamel.yaml

import FAIRsoft
from FAIRsoft import utils
import bioconda_utils.recipe as brecipe # this module is installed in env 'bioconda'
import dotenv

dotenv.load_dotenv()


class Error(Exception):
    """Base class for other exceptions"""
    pass

def run_host_target(tool, type_):
    if tool['requirements'].get('host'):
        for host in tool['requirements']['host']:
            if True in [lang in host for lang in ['r-base', 'python','perl'] ]:
                if tool['requirements'].get('run'):
                    for run in tool['requirements']['run']:
                        if True in [lang in run for lang in ['r-base', 'python','perl'] ]:
                            type_.add('lib')
    return(type_)
    


def get_type(tool):
    '''
    If bioconductor in name and urls:
        Lib
    If test:
        If test.imports: 
            Lib
        If test.commands: 
            CMD
        If source_files:
            CMD
    Else:
        If build.entrypoints: 
            CMD
        Else-If (python OR perl OR r-base) in requirements.host:
            If (python OR perl OR r-base) in requirements.run:
                Lib
        Else:
            CMD
    '''
    type_ = set()
    bioconductor = False

    if 'bioconductor-' in tool['name']:
        bioconductor = True
    if tool.get('source'):
        # source can be a list or a dict. The following handles it

        if isinstance(tool.get('source'), ruamel.yaml.comments.CommentedMap):
            source = [tool.get('source')]    
        elif isinstance(tool.get('source'), ruamel.yaml.comments.CommentedSeq):
            source = tool.get('source')
        else:
            source = []
            #print(type(tool.get('source')))
            #print(len(tool.get('source')))

        # source = list(s1, s2, ....)
        for l in source:
            if l.get('url'):
                for url in l['url']:
                    if 'bioconductor' in url:
                        bioconductor = True

    if tool.get('about'):
        if tool['about'].get('home'):
            if 'bioconductor' in tool['about']['home']:
                bioconductor = True

    if bioconductor == True:
        type_.add('lib')

    if tool.get('test'):
        if 'commands' in tool['test']:
            if tool['test']['commands']:
                if True not in ['$R' in command for command in tool['test']['commands']]:
                    type_.add('cmd')
        if 'imports' in tool['test']:
            if tool['test']['imports']:
                type_.add('lib')
        if 'source_files' in tool['test']:
            if tool['test']['source_files']:
                type_.add('cmd')

    else:
        if tool.get('build'):
            if tool['build'].get('entrypoints'):
                type_.add('cmd')
        elif tool.get('requirements'):
            type_ = run_host_target(tool, type_)
            
    if len(type_) == 0:
        type_.add('cmd')

    return(type_)

def build_id(tool):
    id_template = "https://openebench.bsc.es/monitor/tool/bioconda_recipes:{name}:{version}/{type}"
    name = tool['package']['name']
    type_= list(get_type(tool))
    tool['type'] = type_
    id_ = id_template.format(name=tool['name'], version=None, type=type_)
    return(id_)

def get_tool_names(subdirectories):
    tool_names_subs = []
    tool_names_subs_raw = []
    bioconductors = 0
    for directory in subdirectories:
        name = directory.split('/')[-2]
        tool_names_subs_raw.append(name)
    return(tool_names_subs_raw)

class NonPackageError(Error):
    """Raised when a meta.yml is not a package. Usually generated by other recipes"""
    pass

def extract_metadata(package, recipes_path, log):
    insts = []
    try:
        recipe = brecipe.load_parallel_iter(recipes_path, package)
    except Exception as e:
        log['errors'].append({'package':package,'error':e})
        return(insts, log)
    else:
        for a in recipe:
            a.render
            a.meta['name'] = package
            if a.meta.get('package'):
                if a.meta['package'].get('name'):
                    a.meta['name']=a.meta['package']['name']
            
            insts.append(a.meta)
        
        return(insts, log)


def retrieve_packages_metadata(tool, recipes_path, log):
    inst_dicts, log = extract_metadata(tool,recipes_path, log)
    
    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')
    ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique')

    if STORAGE_MODE =='db':
        alambique = FAIRsoft.utils.connect_collection(ALAMBIQUE)
    else:
        OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
        OUTPUT_BIOCONDA_RECIPES = os.getenv('OUTPUT_BIOCONDA_RECIPES', 'bioconda_recipes.json')
        output_file = OUTPUT_PATH + '/' + ALAMBIQUE + '/' + OUTPUT_BIOCONDA_RECIPES

    if inst_dicts:
        for inst_dict in inst_dicts:
            if inst_dict:
                inst_dict['@id'] = build_id(inst_dict)  
                inst_dict['@data_source'] = 'bioconda_recipes'
                if STORAGE_MODE=='db':
                    try:
                        updateResult = alambique.update_one({'@id':tool['@id']}, { '$set': tool }, upsert=True)
                    except Exception as e:
                        log['errors'].append({'file':tool,'error':e})
        
                    else:
                        log['n_ok'] += 1
                else:
                    log = FAIRsoft.utils.save_entry(inst_dict, output_file, log)
    else:
        log['errors'].append({'file':tool,'error':'Empty inst_dict'}) 
    return(log)

def print_progress(log):
    # Keeping track of progress
    #print(f"{len(log['names'])} packages processes --- {log['n_ok']} parsed and loaded sucessfully --- {len(log['errors'])} raised an exception", end="\r", flush=True)
    pass

def process_recipes():
    # 3. List tool names in the directory
    try:
        recipes_path = os.getenv('RECIPES_PATH')        
        print(recipes_path)
    except:
        raise Exception('RECIPES_PATH environment variable not set. Exiting importation.')
    else:
        subdirectories = glob("%s/*/"%(recipes_path))
        tool_names_subs_raw = get_tool_names(subdirectories)
        print('List of names obtained')
        print('Number of tools: %s'%(len(tool_names_subs_raw)))
    
        # For each tool
        print('Processing metadata')
        log = {'names':[],
                'n_ok':0,
                'errors': []}
        for tool in tool_names_subs_raw:
            # 4. Process and push metadata
            print(tool)
            log = retrieve_packages_metadata(tool, recipes_path, log)
            print_progress(log)

    
    print(f"----- Importation finished ----- \n Number of packages in Bioconda {len(log['names'])}")
    print('Exceptions\n')
    for e in log['errors']:
        print(e['error'])  
    


if __name__=='__main__':
    # Extract and push metadata
    process_recipes()





from .importer_recipes import process_recipes


def import_data():
    # ------- 1. recipes -----------------------
    # 1.1 Set recipes' path
    process_recipes()

if __name__=='__main__':
    import_data()



import bioconductor_imp.importer as bioconductor
import bioconda_imp.importer as bioconda
import bioconda_biotools_galaxy_imp.importer as opeb_tools
import toolshed_imp.importer as toolshed
import repositories_imp.importer as repos
import opeb_metrics_imp.importer  as opeb_metrics
import sourceforge_imp.importer as sf

def execute_importation(label):
    '''
    Decorator to inform when importation stats and finishes
    '''
    def decorator(func):
        def wrapper():
            print(f'{label} importation starting...')
            func()
            print(f"{label} importation done")
        return wrapper
    return decorator

if __name__ == '__main__':
    # Runs all importators one after another

    importers = [
        bioconductor,
        bioconda,
        opeb_tools,
        toolshed,
        repos,
        opeb_metrics,
        sf
        ]
    
    for imp in importers:
        execute_importation(str(imp))(imp.import_data())



import sys
import configparser
from munch import munchify
from pymongo import MongoClient

from FAIRsoft.utils import instance


def  connect_to_db(config_file_path):
    # connecting to DB
    config = configparser.ConfigParser()
    config.read(config_file_path)
    DBHOST = config['MONGO_DETAILS']['DBHOST']
    DBPORT = config['MONGO_DETAILS']['DBPORT']
    DATABASE = config['MONGO_DETAILS']['DATABASE']
    TOOLS = config['MONGO_DETAILS']['TOOLS']
    FAIR = config['MONGO_DETAILS']['FAIR']

    connection = MongoClient(DBHOST, int(DBPORT))

    tools = connection[DATABASE][TOOLS]
    fair = connection[DATABASE][FAIR]
    return(tools, fair)


def prepFAIRcomp(instances):
    stdFormats= getFormats(instances)
    return(stdFormats)
    

def getFormats(instances):
    inputs = [a.input for a in instances]
    inputs_ = [a for a in inputs]
    inputsNames = []

    nonSFormats = ['txt', 'text', 'csv', 'tsv', 'tabular', 'xml', 'json', 'nucleotide', 'pdf', 'interval' ]
    for List in inputs_:
        for eachD in List:
            if 'format' in eachD.keys():
                if ' format' not in eachD['format']['term'] and eachD['format']['term'].lstrip() not in nonSFormats:
                    if '(text)' not in eachD['format']['term']:
                        if eachD['format']['term'].lstrip() not in inputsNames:
                            inputsNames.append(eachD['format']['term'].lstrip())
    return(inputsNames)


def convert_dict2instance(tool):
    NewInst = instance(tool['name'], tool['type'], tool['version'])
    NewInst.__dict__ = munchify(tool)
    NewInst.set_super_type()

    return(NewInst)

def computeFAIR(instances, stdFormats):
    for ins in instances:
        ins.generateFAIRMetrics(stdFormats)
        ins.FAIRscores()


def build_indicators_scores(instances):
    print('Saving indicators and scores')
    out_inst_metrics_scr = []
    for ins in instances:
        dic = { **ins.metrics.__dict__, **ins.scores.__dict__ }
        # name, version, type are needed to identify the instance
        dic['name'] = ins.name
        dic['type'] = ins.type
        dic['version'] = ins.version
        out_inst_metrics_scr.append(dic)

    print("Metrics and scores saved")
    return(out_inst_metrics_scr)   

def push_to_db(indicators_scores, fair):
    print('Pushing to DB')
    fair.insert_many(indicators_scores)
    print('Pushed to DB')


def computeScores(tools):
    instances = []

    for tool in tools.find():
        Inst = convert_dict2instance(tool)
        instances.append(Inst)

    global stdFormats
    stdFormats = prepFAIRcomp(instances)

    print('All dicts converted to instances')
    prepFAIRcomp(instances)

    print('Computing indicators and scores ...')
    computeFAIR(instances, stdFormats)

    print("Building objects of instances' indicators and scores (with instance ID) ...")
    indicators_scores = build_indicators_scores(instances)

    return(indicators_scores)

def computeScores_from_list(tools):
    instances = []

    for tool in tools:
        Inst = convert_dict2instance(tool)
        instances.append(Inst)

    global stdFormats
    stdFormats = prepFAIRcomp(instances)

    print('All dicts converted to instances')
    prepFAIRcomp(instances)

    print('Computing indicators and scores ...')
    computeFAIR(instances, stdFormats)

    print("Building objects of instances' indicators and scores (with instance ID) ...")
    indicators_scores = build_indicators_scores(instances)

    return(indicators_scores)


if __name__=='__main__':
    #get config from cmd parameters
    configfile = sys.argv[1]

    tools, fair = connect_to_db(configfile)
    indicators_scores = computeScores(tools)

    push_to_db(indicators_scores, fair)




import argparse
import json
import requests
import re
import subprocess
import bs4
import os

from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from webdriver_manager.chrome import ChromeDriverManager

from .list_packages import fetch_all_packages
import FAIRsoft
from FAIRsoft import utils


def get_program_arguments(arguments_parser): 
    arguments = arguments_parser.parse_args()
    return(arguments)

def program_arguments_parser():
    parser = argparse.ArgumentParser(
            prog='',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--opeb', dest='opeb', action='store_true', help='If tools for which metadata is retrieved are those in OPEB (enricher mode). Otherwise all tools in Bioconductor (putre importer option). Default: all.')
    parser.add_argument('-urls_file_path', required=False, help='''Path file containing the list of URLs to scrap for enricher mode.''')
    return(parser)

# initializing session
session = requests.Session()
headers = {"User-Agent":"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit 537.36 (KHTML, like Gecko) Chrome",
"Accept":"text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8"}

# the html retriever
def getHTML(url, verb=False):
    '''
    Takes and url as an input and returns the corresponding
    bs4 object
    '''
    from bs4.dammit import EncodingDetector
    try:
        re = session.get(url, headers=headers, timeout=(10, 30))
    except:
        print(r'Cannot access HTML')
        return(None)
    else:
        if re.status_code == 200:
            # dealing with encoding
            http_encoding = re.encoding if 'charset' in re.headers.get('content-type', '').lower() else None
            html_encoding = EncodingDetector.find_declared_encoding(re.content, is_html=True)
            encoding = html_encoding or http_encoding

            # generating BeautifulSoup object
            bsObj = BeautifulSoup(re.content, 'html5lib', from_encoding=encoding)

            if verb == True:
                print("The title of html is %s"%bsObj.title.getText())
            return(bsObj)
        else:
            return(None)


def preParsing(bs):
    '''
    This function return to parts of the bsObject.
    part 1: extended name, description, authors, maintainers and citation
    part 2: a dictionary with all fragments under a <h3>.
    '''
    Name = bs.find('h1').get_text()
    html1 = []
    html2 = {}
    htmlsub2 = []
    for tag in bs.find('h2').next_siblings:
        if tag.name == 'h3':
            name = tag.contents[0]
            break
        else:
            html1.append(tag)

    if bs.find('div', attrs={"class":"bioc_citation"}):
        citation = 1
    else:
        citation = 0

    # Getting second part
    for tag in bs.find('h3').next_siblings:
        if tag.name == 'h3':
            html2[name] = htmlsub2
            htmlsub2 = []
            name = tag.contents[0]
        else:
            htmlsub2.append(tag)
    html2[name] = htmlsub2
    return(Name, citation, html1, html2)

def getInsDocDetArch(part2):
    instRaw = part2.get('Installation', None)
    docRaw = part2.get('Documentation', None)
    detailsRaw = part2.get('Details', None)
    packRaw = part2.get('Package Archives', None)

    return(instRaw, docRaw, detailsRaw, packRaw)


def parsePart1(bs, items):
    ps = [p for p in bs if p.name == 'p' ]
    if 'Bioconductor version' in ps[0].contents[0]:
        items['description'] = ps[1].contents[0]
        Auth = [a for a in ps[2].contents[0].split(':')[1].split(',') ]
        if ' and ' in Auth[-1]:
            last_auth = Auth[-1].split('and')
            items['authors'] = Auth[:-1] + last_auth
        else:
            items['authors'] = Auth

        Mant = [a for a in ps[3].contents[0].split(':')[1].split(',') ]
        if ' and ' in Mant[-1]:
            last_mant = Mant[-1].split('and')
            items['mantainers'] = Mant[:-1] + last_mant
        else:
            items['mantainers'] = Mant

    else:
        print('Unknown structure of part1. Aborting the parsin...')
    return(items)


def existInstallation(bs, items):
    instr = [a for a in bs if a.name == 'pre']
    if len(instr)>0:
        items['Installation instructions'] = True
    else:
        items['Installation instructions'] = False

    return(items)


def parseDocumentation(bs, items):
    table = [a for a in bs if a.name == 'table']
    if len(table) > 1:
        print('error in documentation parsing, aborting...')
    else:
        table = table[0]

    items['documentation'] = {}

    for row in table.findAll('tr'):
        cells = [cell for cell in row.findAll('td')]
        if cells[0].find('a') or cells[1].find('a'):
            doc = str(cells[0].find('a'))
            script = str(cells[1].find('a'))
            field = cells[2].get_text()
            field = field.replace('.','_')
            items['documentation'][field] = [doc, script]
    return(items)


def parseDetails(bs, items):
    table = [a for a in bs if a.name == 'table']
    if len(table) > 1:
        print('error in details parsing, aborting...')
    else:
        table = table[0]

    for row in table.findAll('tr'):
        cells = [cell for cell in row.findAll('td')]
        if re.sub('\s+', '', cells[1].get_text() ) in ['', ' ']:
            items[cells[0].get_text()] = None
        else:
            items[cells[0].get_text()] = re.sub('\s+', '', cells[1].get_text() )

    return(items)

import re
def parseArchives(bs, items):
    table = [a for a in bs if a.name == 'table']
    if len(table) > 1:
        print('error in archives parsing, aborting...')
    else:
        table = table[0]

    for row in table.findAll('tr'):
        cells = [cell for cell in row.findAll('td')]
        if re.sub('\s+', '', cells[1].get_text() ) in ['', ' ']:
            items[cells[0].get_text()] = None
        else:
            items[cells[0].get_text()] = re.sub('\s+', '', cells[1].get_text() )

    return(items)

def parseInput(InputPath):
    '''
    This function takes the input file and builds a list of urls
    required: the input path. Line format in input file: <url>\t<name>\n
    returns: a lis of urls in the file
    '''
    urls_file = open(InputPath, 'r')
    urls = []
    counter = 1
    counter_valid = 0
    for line in urls_file:
        if len(line.split('\t')) != 2:
            print("Input file line %s skipped: impossible to parse, number of columns != 2."%(counter))
        else:
            url = line.split('\t')[0]
            if "galaxy" not in url and "github" not in url:
                urls.append(check_protocol(url))
                counter_valid += 1
        counter += 1
    print("Number of URLs to be analyzed: %s"%(counter_valid))
    if urls == []:
        print("No URLs to analyze")
        return(None)
    else:
        return(urls)


def parseInput(InputPath):
    '''
    This function takes the input file and builds a list of urls
    required: the input path. Line format in input file: <url>\t<name>\n
    returns: a lis of urls in the file
    '''
    urls_file = open(InputPath, 'r')
    urls = []
    counter = 1
    counter_valid = 0
    for line in urls_file:
        if len(line.split('\t')) != 2:
            print("Input file line %s skipped: impossible to parse, number of columns != 2."%(counter))
        else:
            url = line.split('\t')[0]
            match = re.search("^(https?:\/\/)?(www.)?bioconductor.org\/packages\/[A-Za-z0-9_.]{1,15}\/bioc\/html\/.*?$", url)
            if match:
                urls.append(check_protocol(url))
                counter_valid += 1
        counter += 1
    print("Number of URLs to be analyzed: %s"%(counter_valid))
    if urls == []:
        print("No URLs to analyze")
        return(None)
    else:
        return(urls)


def check_protocol(url):
    '''
    '''
    if re.match('^http', url) != None: # To be changed for proper regex (beginning of line)
        return(url)
    else:
        return('https://' + url)

def clean_fields(items):
    #Split lists
    list_fields = ['Depends on me', 'Suggests', 'Suggests Me', 'Depends']
    for field in list_fields:
        if field in items and items[field]:
            items[field] = items[field].split(',')
    # remove dots
    keys=tuple(items.keys())
    for key in keys:
        if '.' in key:
            items[key.replace('.', '_')] = items[key]
            items.pop(key, None)
    return(items)


def parse_citation(citation):
    with open('tmp_citation.txt', 'w') as file_name:
        file_name.write(citation)

    command = ["anystyle","--stdout","-f", "json","parse","tmp_citation.txt"]
    process = subprocess.run(args = command, stdout=subprocess.PIPE).stdout
    #parsed = ast.literal_eval(process.decode("utf-8"))
    parsed = json.loads(process.decode("utf-8"))
    if parsed:
        return(parsed[0])
    else:
        return([])

import ast
def get_publication(driver, url, items):
    driver.get(url)
    try:
        citation = driver.find_element_by_class_name('bioc_citation')
    except:
        citation = None
        link = None
    finally:
        try:
            link = citation.find_element_by_tag_name('a')
            link = link.get_attribute('href')
        except:
            link = None

    pub_cit = parse_citation(citation.text)
    pub = {"url":link, "citation": pub_cit}
    items["publication"] = pub
    return(items)

def retrieve_metadata(bs, url, driver):
    items = {}
    name, citation, part1, part2 = preParsing(bs)
    items['name'] = name
    items['citation'] = citation
    instRaw, docRaw, detailsRaw, packRaw = getInsDocDetArch(part2)
    items = parsePart1(part1, items)
    items = existInstallation(instRaw, items)
    items = parseDocumentation(docRaw, items)
    items = parseDetails(detailsRaw, items)
    items = parseArchives(packRaw, items)

    if items['URL'] != None:
        items['links'] = [url] + [items['URL']]
    else:
        items['links'] = [url]
    
    items = get_publication(driver, url, items)
    items = clean_fields(items)
    items['@data_source'] = 'bioconductor'
    items['@source_url'] = url
    items['@id'] = f"https://openebench.bsc.es/monitor/tool/bioconductor:{name}:{items['Version']}/lib"
    return(items)


def import_data():
    # 1. connect to DB/ get output file
    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')
    ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique')
    print('hello, this is the bioconductor importer')

    if STORAGE_MODE =='db':
        alambique = FAIRsoft.utils.connect_collection(ALAMBIQUE)
    else:
        OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
        OUTPUT_BIOCONDUCTOR = os.getenv('OUTPUT_BIOCONDUCTOR', 'bioconductor.json')
        output_file = OUTPUT_PATH + '/' + ALAMBIQUE + '/' + OUTPUT_BIOCONDUCTOR
        print(output_file)

    
    # 2. getting arguments
    driver = webdriver.Chrome(ChromeDriverManager().install())
    arguments_parser = program_arguments_parser()
    arguments = get_program_arguments(arguments_parser)
    
    # 3. Fetch packages
    if arguments.opeb == True:
        urls_file = arguments.urls_file_path
        urls = parseInput(urls_file)
    else:
        urls = fetch_all_packages()

    log = {'names':[],
                'n_ok':0,
                'errors': []}
    if urls:
        n = 0
        print("Requesting packages URLs...")
        for url in urls:
            print("Package %d of %d"%(n, len(urls)), end='\r')
            bsO = getHTML(url, verb=False)
            n+=1
            if bsO != None:
                print('Processing BeautifulSoup Objects')
                inst_dict = retrieve_metadata(bsO, url, driver)
                print('Saving data') 

                if STORAGE_MODE=='db':
                    log = FAIRsoft.utils.push_entry(inst_dict, alambique, log)
                else:
                    print(output_file)
                    log = FAIRsoft.utils.save_entry(inst_dict, output_file, log)

    else:
        print('No URLS found')
        


if __name__ == '__main__':
    import_data()


import json
import re
import logging
import bibtexparser

from FAIRsoft.utils import setOfInstances
from FAIRsoft.utils import instance

class toolGenerator(object):
    def __init__(self, tools, source):
        self.tools = tools
        self.source = source
        logging.debug('Generator for ' + self.source + ' initialized') 

# --------------------------------------------
# Functions used by several transformations
# --------------------------------------------

def clean_version(version):
    if version != None:
        if '.' in version:
            return(version.split('.')[0]+'.'+ version.split('.')[1])
        else:
            return(version)
    else:
        return(version)


def clean_name(name):
    bioconductor=re.search("^bioconductor-", name)
    if bioconductor:
        name=name[bioconductor.end():]
    emboss_dots=re.search("^emboss: ", name)
    if emboss_dots:
        name=name[emboss_dots.end():]
    emboss_unders=re.search("^emboss__", name)
    if emboss_unders:
        name=name[emboss_unders.end():]
    return(name)


def extract_ids(id_):
    #extract ids from metrics @id
    fields = id_.split('/')
    if len(fields)>6:
        name = fields[5].split(':')[1]
        if len(fields[5].split(':'))>2:
            version = fields[5].split(':')[2]
            source = fields[5].split(':')[0]
        else:
            version = None
            source = fields[5].split(':')[0]
        type_ = fields[6]
    
        ids = {
            'name' : name,
            'version' : version,
            'type' : type_,
            'source': source
        }

        return(ids)
    
    return(False)


def get_repo_name_version_type(id_):
    fields = id_.split('/')
    name_plus = fields[5]
    name_plus_fields=name_plus.split(':')
    name=name_plus_fields[1]
    if len(name_plus_fields)>2:
        version=name_plus.split(':')[2]
    else:
        version=None 

    if len(fields)>6:
        type_=fields[6]
    else:
        type_=None
     
    return({'name':name, 'version':version, 'type':type_})

# --------------------------------------------------------------
#   TRANSFORMERS
# --------------------------------------------------------------

# --------------------------------------------------------------
# Repository transformer 
# --------------------------------------------------------------
class repositoryToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'repository'):
        toolGenerator.__init__(self, tools, source)
        self.instSet = setOfInstances('repository')
        self.import_bioconda_dict()

        for tool in self.tools:
            # We skip generic entries
            if len(tool['@id'].split('/'))<7:
                continue

            else:
                id_data = get_repo_name_version_type(tool['@id'])
                name = clean_name(id_data['name'].lower())
                version = id_data['version']
                if version == None:
                    version = 'unknown'

                types_=[id_data['type']]

                if self.bioconda_types.get(name):
                    types_ = self.bioconda_types[name]
                
                if id_data['type'] == 'workflow' and 'galaxy' in tool['@id']:
                    types_=['cmd']

                for type_ in types_:

                    newInst = instance(name, type_, [version])
                    # there are several versions, to simplify integration, we want one instance per version
                    if 'versions' in  tool['repos'][0]['res'].keys():
                        versions = tool['repos'][0]['res']['versions']
                    else:
                        versions = [version]

                    for v in versions:
                        newInst = instance(name, type_, versions)
                        
                        newInst.label = clean_name(id_data['name'])

                        if tool['repos'][0]['res'].get('desc'):
                            newInst.description = [tool['repos'][0]['res'].get('desc')]
                        newInst.links = tool['entry_links']
                        newInst.publication =  []

                        binary_uri = tool['repos'][0]['res'].get('binary_uri')
                        source_uri = tool['repos'][0]['res'].get('source_uri')
                        download = [binary_uri, source_uri]
                        if None in download:
                            download.remove(None)
                        if source_uri or binary_uri:
                            newInst.download = download
                        else:
                            newInst.download = []  # list of lists: [[type, url], [], ...]
                        
                        newInst.inst_instr =  tool['repos'][0]['res'].get('has_tutorial')
                        newInst.test = None
                        try:
                            src = [tool['repos'][0]['res'].get('source_uri')]
                        except:
                            newInst.src = []
                        else:
                            if tool['repos'][0]['res'].get('source_uri')!=None:
                                newInst.src = src

                        if newInst.src:
                            newInst.os = ['Linux', 'Mac', 'Windows']
                        else:
                            newInst.os = None

                        newInst.input = None # list of dictionaries biotools-like {'format' : <format> , 'uri' : <uri> , 'data' : <data> , 'uri': <uri>}
                        newInst.output = None  # list of dictionaries biotools-like {'format' : <format> , 'uri' : <uri> }
                        newInst.dependencies = None # list of strings
                        if tool['repos'][0]['res'].get('readmeFile'):
                            newInst.documentation = [['readme', tool['repos'][0]['res'].get('readmeFile')]] # list of lists [[type, url], [type, rul], ...]
                        if tool['repos'][0]['res'].get('source_license'):
                            newInst.license = [tool['repos'][0]['res'].get('source_license')] 
                            newInst.termsUse = [tool['repos'][0]['res'].get('source_license')] 
                        newInst.contribPolicy = None

                        auths = []
                        authors_l = tool['repos'][0]['res'].get('tool_developers')
                        if authors_l:
                            for author in authors_l:
                                auths.append(author.get('username', None))
                        newInst.authors = auths

                        newInst.repository = tool['entry_links']
                        newInst.source = [tool['repos'][0]['kind']] #string
                        newInst.bioschemas = None
                        newInst.https = None
                        newInst.operational = None
                        newInst.ssl = None
                
                        self.instSet.instances.append(newInst)
    

    def import_bioconda_dict(self):
        '''
        TODO: the path to bioconda_dict should be a parameter
        '''
        path = '/Users/evabsc/projects/FAIRsoft/FAIRsoft_ETL/FAIRsoft/FAIRsoft/transformation/bioconda_types.json'
        logging.info('Importing bioconda types from: ' + path)
        with open(path, 'r') as infile:
            self.bioconda_types = json.load(infile)


# --------------------------------------------------------------
# Bioconda recipes tranformer 
# --------------------------------------------------------------
class biocondaRecipesToolsGenerator(toolGenerator):
    def __init__(self, tools, source="bioconda_recipes"):
        toolGenerator.__init__(self, tools, source)
        self.instSet = setOfInstances('bioconda_recipes')

        for tool in self.tools:
            name = clean_name(tool['name']).lower()
            version = tool['@id'].split('/')[5].split(':')[2]
            #print(tool)
            types_ = tool['type']
            if not types_:
                types_ =  ['cmd']

            if version == None:
                version = None
            
            for type_ in types_:
            
                newInst = instance(name, type_, [version])
                try:
                    description = tool['about']['description']
                except:
                    newInst.description = []
                else:
                    if description:
                        newInst.description = [description] # string
                
                newInst.label = clean_name(tool['name'])

                links = []
                if 'about' in tool.keys() and tool['about']:
                    if 'home' in tool['about'].keys() and tool['about']['home']:
                        if tool['about']['home']:
                            links.append(tool['about']['home'])
            
                src = []
                if tool.get('source'):
                    if type(tool.get('source')) == dict:
                        tool['source'] = [tool['source']]
                    for block in tool['source']:
                        if block.get('url'):
                            src = block['url']
                            if type(block['url'])==list:
                                for l in block['url']:
                                    links.append(l)

                            else:
                                links.append(block['url'])
                
                try:
                    doc_url = tool['about']['doc_url']
                except:
                    pass
                else:
                    if doc_url:
                        for d in doc_url:
                            links.append(d)

                # publications
                newInst.links = links

                new_pubids = set()
                if tool.get('extra'):
                    if tool['extra'].get('identifiers'):
                        for ident in tool['extra'].get('identifiers'):
                            reg1 = r'https:\/\/doi.org\/(10.([\w.]+?)\/([\w.]+)([\w.\/]+)?)'
                            reg2 = r'doi:(10.([\w.]+?)\/([\w.]+)([\w.\/]+)?)'
                            m1 = re.match(reg1, ident)
                            m2 = re.match(reg2, ident)
                            if m1:
                                new_pubids.add(m1.group(1))
                            if m2:
                                new_pubids.add(m2.group(1))

                    if tool['extra'].get('doi'):
                        for doi in tool['extra'].get('doi'):
                            new_pubids.add(doi)
                
                new_pubs = []
                for unique_id in new_pubids:
                    new_pubs.append({'doi':unique_id})

                newInst.publication =  new_pubs
                if src:
                    if type(src) != list:
                        src = [src]
                    
                    newInst.download = src
                    
                newInst.inst_instr = True # boolean // FUTURE: u'ri or text
                newInst.test = None
                newInst.src = src # string
                newInst.os = ['Linux', 'Mac', 'Windows'] # list of strings
                newInst.input = [] # list of dictionaries biotools-like {'format' : <format> , 'uri' : <uri> , 'data' : <data> , 'uri': <uri>}
                newInst.output = [] # list of dictionaries biotools-like {'format' : <format> , 'uri' : <uri> }
                deps = [] 
                if 'requirements' in tool.keys() and tool['requirements']:
                    for req_k in tool['requirements'].keys():
                        if tool['requirements'][req_k]:
                            for dep in tool['requirements'][req_k]:
                                deps.append(dep)

                newInst.dependencies = deps # list of strings
                documentation = []
                try:
                    doc = ['documentation',tool['about']['docs']]
                    documentation.append(doc)
                except:
                    pass

                try:
                    doc = ['documentation', tool['about']['doc_url']]
                    documentation.append(doc)
                except:
                    pass

                newInst.documentation = documentation # list of lists [[type, url], [type, rul], ...]
                if 'about' in tool.keys():
                    if tool['about'].get('license'):
                        newInst.license = [tool['about'].get('license')] 
                        newInst.termsUse = [tool['about'].get('license')] #
                newInst.contribPolicy = False
                credit = []
                try:
                    auth=tool['about']['author']
                except:
                    pass
                else:
                    credit.append(auth)
                try:
                    mantainers = tool['about']['mantainers']
                except:
                    mantainers = None
                else:
                    for m in mantainers:
                        credit.append(m)

                newInst.authors = credit # list of strings
                repository = set()
                for l in links:
                    if l:
                        if 'github' in l:
                            reg = r'^(https?:\/\/)?(www)?\b(github.com)\/[^/]*\/[^/]*'
                            m = re.match(reg, l)
                            if m:
                                repository.add(m.group())
                        #TODO: clean remaining repo extraction
                        elif 'bitbucket' in l:
                            repository.add(l)
                        elif 'sourceforge' in l:
                            repository.add(l)

                newInst.repository = list(repository)
                newInst.source = ['bioconda_recipes'] #string
                newInst.bioschemas = None
                newInst.https = None
                newInst.operational = None
                newInst.ssl = None

                self.instSet.instances.append(newInst)
    

    def type_dictionary(self, outpath):
        instances_types = {}
        for inst in self.instSet.instances:
            if inst.name in instances_types:
                instances_types[inst.name].append(inst.type)
            else:
                instances_types[inst.name] = [inst.type]
        
        if outpath:
            with open(outpath, 'w') as out_file:
                json.dump(instances_types, out_file)

        return(instances_types)

#--------------------------------------------
# Bioconda Conda Transformer
#--------------------------------------------

class bioconda_conda_ToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'bioconda_conda'):
        toolGenerator.__init__(self, tools, source)
        self.import_bioconda_dict()
        self.instSet = setOfInstances('bioconda_conda')

        for tool in self.tools:
            id_info = extract_ids(tool['@id'])
            name = clean_name(id_info['name']).lower()
            version = id_info['version']
            if version == None:
                version = 'unknown'
    
            if self.bioconda_types.get(name):
                types_ = self.bioconda_types[name]
            else:
                types_ = ['cmd']

            for type_ in types_:
                newInst = instance(name, type_, [version])

                newInst.label = clean_name(id_info['name']) # string
                newInst.description = None # string
                newInst.version = [version]
                newInst.type = type_
                newInst.links =[tool['url']]
                newInst.publication =  [] # number of related publications [by now, for simplicity]
                newInst.download = [tool['url']]  # list of lists: [[type, url], [], ...]
                newInst.inst_instr = True # boolean // FUTURE: uri or text
                newInst.test = None # boolean // FUTURE: uri or text
                newInst.src = [] # string
                newInst.os = ['Linux', 'Mac', 'Windows'] # list of strings
                newInst.input = [] # list of dictionaries biotools-like {'format' : <format> , 'uri' : <uri> , 'data' : <data> , 'uri': <uri>}
                newInst.output = [] # list of dictionaries biotools-like {'format' : <format> , 'uri' : <uri> }
                newInst.dependencies = tool['dependencies'] # list of strings
                newInst.documentation = [] # list of lists [[type, url], [type, rul], ...]
                if 'license' in tool.keys():
                    newInst.license = [tool['license']] # string
                newInst.termsUse = None #
                newInst.contribPolicy = None
                newInst.authors = [] # list of strings
                newInst.repository = []
                newInst.source = ['bioconda_conda'] #string
                newInst.bioschemas = None
                newInst.https = None
                newInst.operational = None
                newInst.ssl = None

                self.instSet.instances.append(newInst)

    def import_bioconda_dict(self):
            with open('./bioconda_types.json', 'r') as infile:
                self.bioconda_types = json.load(infile)

# --------------------------------------------
# Bioconda Tools Transformer
# --------------------------------------------
class biocondaToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'bioconda'):
        toolGenerator.__init__(self, tools, source)

        self.import_bioconda_dict()
        self.instSet = setOfInstances('bioconda')

        #names = [a['name'].lower() for a in self.tools]
        #print('diferent names in bioconda tools: ' + str(len(set(names))))

        for tool in self.tools:
            name = clean_name(tool['@label']).lower()
            
            version = clean_version(tool['@version'])
            if version == None:
                version = 'unknown'
            
            if self.bioconda_types.get(name):
                    types_ = self.bioconda_types[name]
            else:
                types_ = ['cmd']
                
            for type_ in types_:
                newInst = instance(name, type_, [version])

                newInst.label = clean_name(tool['@label']) # string

                if 'description' in tool.keys():
                    newInst.description = [tool['description']] # string
                if tool['web']['homepage']:
                    newInst.links = [tool['web']['homepage']] #list
                else:
                    newInst.links = []
                
                #print(tool['publications'] )
                newInst.publication =  tool['publications'] 

                download = []
                for k in tool['distributions'].keys():
                    for link in tool['distributions'][k]:
                        download.append(link)

                newInst.download = download

                newSrc = []
                for down in tool['distributions'].keys():
                    if 'source' in down:
                        if len(tool['distributions'][down])>0:
                            for u in tool['distributions'][down]:
                                newSrc.append(u)
                newInst.src = newSrc 

                if 'license' in tool.keys() and tool['license']!='':
                    newInst.license = [tool['license']] # string
                newInst.repository = tool['repositories']
                newInst.source = ['bioconda']
                newInst.links.append(tool['web']['homepage'])
                if tool['repositories']:
                    for a in tool['repositories']:
                        newInst.links.append(a)

                self.instSet.instances.append(newInst)
        

    def import_bioconda_dict(self):
            with open('/Users/evabsc/projects/FAIRsoft/FAIRsoft_ETL/FAIRsoft/FAIRsoft/transformation/bioconda_types.json', 'r') as infile:
                self.bioconda_types = json.load(infile)


def constrFormatsConfig(formatList):
    '''
    From an input that is a str to a biotools kind of format
    '''
    notFormats = ['data']
    newFormats = []
    seenForms = []
    for formt in formatList:
        if formt not in seenForms:
            if ',' in formt:
                formats = formt.split(',')
                for f in formats:
                    if f not in notFormats:
                        newFormats.append({ 'format' : {'term' : f , 'uri' :  None }})
                        seenForms.append(formt)
            else:
                if formt not in notFormats:
                    newFormats.append({ 'format' :  {'term' : formt , 'uri' :  None }})
                    seenForms.append(formt)
        else:
            continue
    return(newFormats)


# --------------------------------------------
# Biotools OPEB Tools Transformer
# --------------------------------------------

class biotoolsOPEBToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'biotoolsOPEB'):
        toolGenerator.__init__(self, tools, source)

        self.import_bioconda_dict()
        self.instSet = setOfInstances('biotoolsOPEB')

        for tool in self.tools:
            if tool['@label']:
                name = clean_name(tool['@label']).lower()
                type_ = tool['@type']
                version = clean_version(tool['@version'])

                if self.bioconda_types.get(name):
                    types_ = self.bioconda_types[name]
                elif type_ == 'workflow' and 'galaxy' in tool['@id']:
                    types_=['cmd']
                else:
                    types_ = [type_]

                for type_ in types_:
                
                    if version == None:
                        version = 'unknown'
                    if type_ == None:
                        type_ = 'unknown'

                    newInst = instance(name, type_, [version])

                    newInst.label = clean_name(tool['name']) # string

                    newInst.description = [tool['description']] # string

                    newInst.publication = tool['publications']
                    newInst.test = False
                    if 'license' in tool.keys():
                        newInst.license = [tool['license']]
                    newInst.input = []
                    newInst.output = []
                    if 'documentation' in tool.keys():
                        if 'general' in tool['documentation'].keys():
                            newInst.documentation = [['general', tool['documentation']['general']]]
                    newInst.source = ['biotools']
                    os = []
                    if 'os' in tool.keys():
                        for o in tool['os']:
                            os.append(o)
                        newInst.os = os
                    newInst.repository = tool['repositories']

                    newInst.links.append(tool['web']['homepage'])
                    if tool['repositories']:
                        for a in tool['repositories']:
                            newInst.links.append(a)


                    if tool['semantics']:
                        newInst.input = tool['semantics'].get('inputs', [])
                        newInst.output = tool['semantics'].get('outputs', [])
                        newInst.edam_topics = tool['semantics'].get('topics',[])
                        newInst.edam_operations = tool['semantics'].get('operations',[])
                        newInst.semantics = tool['semantics']

                    newAuth = []
                    for dic in tool['credits']:
                        if dic.get('name'):
                            if dic['name'] not in newAuth and dic['name']!=None:
                                newAuth.append(dic['name'])
                    newInst.authors = newAuth
                    
                    newInst.tags = tool['tags'] 

                    self.instSet.instances.append(newInst)

    def import_bioconda_dict(self):
            with open('./bioconda_types.json', 'r') as infile:
                self.bioconda_types = json.load(infile)


# --------------------------------------------
# SourceForge Transformer
# --------------------------------------------
class sourceforgeToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'sourceforge'):
        toolGenerator.__init__(self, tools, source)

        self.instSet = setOfInstances('sourceforge')

        for tool in self.tools:
            name = tool['@source_url'].split('/')[-1]
            
            name = name.lower()
            type_ = 'unknown'
            version = 'unknown'

            newInst = instance(name, type_, [version])

            newInst.label = tool['name'] # string

            newInst.source = ['sourceforge']
            os = []
            if 'operating_system' in tool.keys():
                for o in tool['operating_system']:
                    os.append(o)
                newInst.os = os
            newInst.repository = [tool['repository']]
            newInst.download = [tool['@source_url']]
            newInst.repository = [tool['@source_url']]
            newInst.links.append(tool['homepage'])
            newInst.description = [tool.get('description')]
            newInst.license = tool['license']

            self.instSet.instances.append(newInst)


# --------------------------------------------
# Galaxy Transformer
# --------------------------------------------
class galaxyOPEBToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'galaxyOPEB'):
        toolGenerator.__init__(self, tools, source)

        self.instSet = setOfInstances('galaxyOPEB')

        for tool in self.tools:
        
            name = clean_name(tool['@label']).lower()
            name=name.replace(' ','_') #so it is coherent with opeb metrics names

            type_ = 'cmd'
            version = tool['@version']
            if version == None:
                version = 'unknown'

            newInst = instance(name, type_, [version])

            newInst.label = clean_name(tool['name']) # string
            newInst.description = [tool['description']] # string
            newInst.source = ['galaxy']
            newInst.os = ['Mac', 'Linux']
            newInst.repository = tool['repositories']
            if 'license' in tool.keys():
                newInst.license = [tool['license']]
            newInst.publication = tool['publications']

            self.instSet.instances.append(newInst)


# --------------------------------------------
# OPEB Metrics Transformer
# --------------------------------------------
class metricsOPEBToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'opeb_metrics'):
        toolGenerator.__init__(self, tools, source)
        self.import_bioconda_dict()
        self.instSet = setOfInstances('opeb_metrics')
        more_pubs = 0
        more_entries = 0
        for tool in self.tools:
            ids = extract_ids(tool['@id'])
            if ids:
                version = ids['version']
                
                # type needs to be corrected for galaxy tools from workflow to web
                
                if version == None:
                    version = 'unknown'
                name = clean_name(ids['name']).lower()

                if self.bioconda_types.get(name):
                    types_ = self.bioconda_types[name]
                elif ids['type'] == 'workflow' and 'galaxy' in tool['@id']:
                    types_=['cmd']
                else:
                    types_ = [ids['type']]
                
                for type_ in types_:
                    newInst = instance(name, type_, [version])
                    newInst.source = ['opeb_metrics']
                    if tool['project'].get('website'):
                        newInst.bioschemas = tool['project']['website'].get('bioschemas')
                        newInst.https = tool['project']['website'].get('https')
                        newInst.ssl = tool['project']['website'].get('ssl')
                        if tool['project']['website'].get('operational') == 200:
                            newInst.operational = True
                        else:
                            newInst.operational = False

                    pubs_new= []
                    if tool['project']['publications']:
                        for pub in tool['project']['publications']:
                            if pub.get('entries'):
                                for entry in pub.get('entries'):
                                    pubs_new.append(entry)
                                if len(pub['entries']) > 1:
                                    more_entries += 1

                        if len(tool['project']['publications'])>1:
                            more_pubs += 1
                        
                        newInst.publication = pubs_new
                    else:
                        newInst.publication = []
                    
                    

                    self.instSet.instances.append(newInst)
        
        logging.debug(f"Entries with more than one publications (entry.project.publication): {more_pubs}")
        logging.debug(f"Publications with more than one entry (entry.project.publication[i].entries) {more_entries}")

    def import_bioconda_dict(self):
        with open('./bioconda_types.json', 'r') as infile:
            self.bioconda_types = json.load(infile)


# --------------------------------------------
# Galaxy Transformer
# --------------------------------------------
class galaxyShedToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'galaxyShed'):
        toolGenerator.__init__(self, tools, source)

        self.instSet = setOfInstances('toolshed')

        for tool in self.tools:
            name = clean_name(tool['name']).lower()
            type_ = 'cmd'
            if 'version' in tool.keys():
                version = clean_version(tool['version'])
            else:
                version = 'unknown'

            newInst = instance(name, type_, [version])

            newInst.label = clean_name(tool['name']) # string
            newInst.description = [tool['description']] # string
            newInst.inst_instr = True # Since this is installable through ToolShed
            if len(tool['tests'])>0:
                newInst.test = True # boolean
            else:
                newInst.test = False

            newInst.dependencies = [a['name'] for a in tool['requirements']] # list of strings
            newInst.repository = [] #
            newInst.links = [] #
            newInst.source = ['toolshed']#string
            newInst.os = ['Linux', 'Mac']

            self.instSet.instances.append(newInst)

# --------------------------------------------
# Galaxy Metadata Transformer
# --------------------------------------------
class galaxyMetadataGenerator(toolGenerator):
    def __init__(self, tools, source = 'galaxy_metadata'):
        toolGenerator.__init__(self, tools, source)

        self.instSet = setOfInstances('galaxy_metadata')

        for tool in self.tools:
            name = clean_name(tool['name']).lower()
            type_ = 'cmd'
            if 'version' in tool.keys():
                version = clean_version(tool['version'])
            else:
                version = 'unknown'

            newInst = instance(name, type_, [version])

            newInst.label = clean_name(tool['name']) # string
            newInst.dependencies = tool['dependencies'] # list of strings
            newInst.source = ['galaxy_metadata']

            self.instSet.instances.append(newInst)

def parse_bibtex(ent):
    parser = bibtexparser.bparser.BibTexParser(common_strings=True)
    logger = logging.getLogger('bibtexparser')
    new_entries = []
    try:
        bibtexdb = bibtexparser.loads(ent, parser=parser)
        for entry in bibtexdb.entries:
            if entry['ENTRYTYPE'].lower() != 'misc':
                #print(entry)
                single_entry = {}
                single_entry['url'] = entry.get('url')
                single_entry['title'] = entry.get('title')
                single_entry['year'] = entry.get('year')
                single_entry['journal'] = entry.get('journal')
                single_entry['doi'] = entry.get('doi')
                single_entry['pmid'] = entry.get('pmid')
                new_entries.append(single_entry)
    except Exception as err:
        logger.error(f'FAILED attempt to parse citation (bibtex). Error: {err}')
        logger.error(json.dumps(ent, sort_keys=False, indent=4))

    return(new_entries)

# --------------------------------------------
# Galaxy Config Transformer
# --------------------------------------------
class galaxyConfigToolsGenerator(toolGenerator):
    def __init__(self, tools, source = 'toolshed'):
        toolGenerator.__init__(self, tools, source)

        self.instSet = setOfInstances('toolshed')

        for tool in self.tools:
            if tool['name']:
                name = clean_name(tool['name']).lower()

                type_ = 'cmd'

                version = clean_version(tool['version'])
                if version == None:
                    version = 'unknown'

                newInst = instance(name, type_, [version])

                newInst.label = clean_name(tool['name']) # string

                if tool['description']:
                    newInst.description = [tool['description']] # string

                new_pubs = []
                if tool['citation']:
                    if tool.get('citation'):
                        for cit in tool['citation']:
                            if cit.get('type') == 'doi':
                                new_pub = {'doi':cit.get('citation')}
                                new_pubs.append(new_pub)
                            else:
                                new_entries = parse_bibtex(cit.get('citation'))
                                for se in new_entries:
                                    new_pubs.append(se)

                    newInst.publication =  new_pubs

                newInst.test = tool['tests'] # boolean

                if len(tool['dataFormats']['inputs'])>0:
                    newInst.input = constrFormatsConfig(tool['dataFormats']['inputs']) # list of strings

                if len(tool['dataFormats']['outputs'])>0:
                    newInst.output = constrFormatsConfig(tool['dataFormats']['outputs']) # list of strings

                docu = []
                if tool['readme'] == True:
                    docu.append(['readme', None])
                if tool['help']:
                    docu.append(['help', tool['help'].lstrip()])

                newInst.documentation = docu # list of lists [[type, url], [type, url], ...]

                newInst.source = ['toolshed'] #string

                self.instSet.instances.append(newInst)



def lowerInputs(listInputs):
    newList = []
    if len(listInputs)>0:
        for format in listInputs:
            newFormat = {}
            for a in format.keys():
                newInner = {}
                if format[a] != []:
                    #print(format[a])
                    if type(format[a]) == list:
                        for eachdict in format[a]:
                            for e in eachdict.keys():
                                newInner[e] = eachdict[e].lower()
                            newFormat[a] = newInner
                    else:
                        for e in format[a].keys():
                            newInner[e] = format[a][e].lower()
                        newFormat[a] = newInner
        newList.append(newFormat)
    else:
        return([])
    return(newList)

# --------------------------------------------
# Bioconductor Transformer
# --------------------------------------------
class bioconductorToolsGenerator(toolGenerator):
    def __init__(self, tools, source='bioconductor'):
        toolGenerator.__init__(self, tools, source)

        self.instSet = setOfInstances('bioconductor')

        for tool in self.tools:
            type_= 'lib'
            version = clean_version(tool['Version'])
            if version == None:
                version = 'unknown'
            name = clean_name(tool['name']).lower()

            newInst = instance(name, type_, [version])

            newInst.label = clean_name(tool['name']) # string
            newInst.description = [tool['description']] # string
            if tool['URL']:
                newInst.links = [tool['URL']]
            else:
                newInst.links = []
            
            newInst.publication = []
            if tool['publication'].get('url'):
                if tool['publication']['citation'].get('type') == 'article-journal':
                    journal = None
                    for a in tool['publication']['citation']['container-title']:
                        if 'ISSN' not in a:
                            journal = a
                    
                    fields = {}
                    for field in ['date', 'title']:
                        if tool['publication']['citation'].get(field):
                            fields[field] = tool['publication']['citation'][field][0]
                        else:
                            fields[field] = None
                    
                    if tool['publication'].get('url'):
                        url = tool['publication']['url'][0]
                    else:
                        url = None
                    
                    newInst.publication =  [{
                    'title': fields['title'],
                    'year': fields['date'],
                    'url': url,
                    'journal': journal,
                    }]

            
            download = []
            for a in ["Windows Binary", "Source Package", "Mac OS X 10 10.11 (El Capitan)"]:
                if a in tool.keys() and tool[a]:
                    download.append(tool['Package Short Url'] + tool[a])

            newInst.download = download
            newInst.inst_instr = tool['Installation instructions'] #
            newInst.src = [ a for a in newInst.download if a[0] == "Source Package"[0] ]# string
            newInst.os = ['Linux', 'Mac', 'Windows'] # list of strings
            if tool['Depends']:
                deps = tool['Depends']
            else:
                deps = []
            if tool['Imports']:
                impo = tool['Imports'].split(',')
            else:
                impo = []

            newInst.dependencies = [item for sublist in [deps+impo] for item in sublist] # list of strings

            newInst.documentation = [[ a, a[0] ] for a in tool['documentation']] # list of lists [[type, url], [type, rul], ...]
            if tool['License']!='':
                newInst.license = [tool['License']] # string
            else:
                newInst.license = False
            newInst.authors = [a.lstrip() for a in tool['authors']] # list of strings
            newInst.repository = [tool['Source Repository'].split('gitclone')[1]]
            newInst.description = [tool['description']]
            newInst.source = ['bioconductor'] #string

            self.instSet.instances.append(newInst)


tool_generators = {
        'bioconductor' : bioconductorToolsGenerator,
        'biotools' : biotoolsOPEBToolsGenerator,
        'bioconda' : biocondaToolsGenerator,
        'toolshed' : galaxyConfigToolsGenerator,
        'galaxy_metadata' : galaxyMetadataGenerator,
        'sourceforge' : sourceforgeToolsGenerator,
        'galaxy' : galaxyOPEBToolsGenerator,
        'opeb_metrics' : metricsOPEBToolsGenerator,
        'bioconda_recipes': biocondaRecipesToolsGenerator,
        'bioconda_conda': bioconda_conda_ToolsGenerator,
        'repository': repositoryToolsGenerator,
}





####
# Old, not used anymore
####
import subprocess
import yaml
import json
import re
import sys
import os
from FAIRsoft import utils
import FAIRsoft  

def build_id(tool):
    id_template = "https://openebench.bsc.es/monitor/tool/bioconda_recipes:{name}:{version}/{type}"
    name = tool['name']
    version = tool.get('version')
    type_=''
    id_ = id_template.format(name=tool['name'], version=version, type=None)
    return(id_)

def retrieve_packages_metadata():
    command = 'conda search -i -c bioconda'
    print('Running command "conda search -i -c bioconda" to get all bioconda packages')
    process = subprocess.check_output(command, shell=True)
    output = str(process)
   
    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')
    ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique') 

    if STORAGE_MODE =='db':
        alambique = FAIRsoft.utils.connect_collection(ALAMBIQUE)
    else:
        OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
        OUTPUT_BIOCONDA_CONDA = os.getenv('OUTPUT_BIOCONDA_CONDA', 'bioconda_conda.json') 
        output_file = OUTPUT_PATH + '/' + ALAMBIQUE + '/' + OUTPUT_BIOCONDA_CONDA

    instances = re.split(r'--[-]*', output)
    log = {'tools':0,
            'n_ok':0,
            'errors': []}
    print('Parsing results...')

    for inst in instances[:10]:
        file_ = ''
        fields = inst.split('\\n')
        for field in fields[1:]:
            field.lstrip
            field = field.replace('::','')
            if ':' in field or '- ' in field:
                file_ = file_ + '\n' + field
        try:
            inst_dict = yaml.safe_load(file_)
        except Exception as e:
            log['errors'].append({'file':file_,'error':e})            
        else:
            if inst_dict:
                if 'conda.anaconda.org/bioconda/' in inst_dict['url']:
                    # 3. Generate id
                    inst_dict['@id'] = build_id(inst_dict)      
                    inst_dict['@data_source'] = 'bioconda_conda'
                    # output to collection/file
                    if STORAGE_MODE=='db':
                        log = FAIRsoft.utils.push_entry(inst_dict, alambique, log)
                    else:
                        log = FAIRsoft.utils.save_entry(inst_dict, output_file, log)

                else:
                    log['errors'].append({'file':file_,'error':'Empty inst_dict'})
            else:
                log['errors'].append({'file':file_,'error':'Empty inst_dict'}) 

        finally:
            log['tools'] += 1
                    
        # Keeping track of progress
        print(f"{log['tools']} packages processes --- {log['n_ok']} parsed and loaded sucessfully --- {len(log['errors'])} failed", end="\r", flush=True)
        #print(log)
    
    # Importation finished
    print(f'''\n----- Importation finished ----- \n Number of packages in Bioconda {log['tools']}''')
    print('Exceptions\n')
    for e in log['errors']:
        print(e['error'])
    
    return
    
if __name__=='__main__':
    # retrieve metadata
    retrieve_packages_metadata()




"""
The command-line interface for the transformer
"""

from .transform_raw import transform
import argparse
import logging
from dotenv import load_dotenv
# assuming loglevel is bound to the string value obtained from the
# command line argument. Convert to upper case to allow the user to
# specify --log=DEBUG or --log=debug

def main():
    parser = argparse.ArgumentParser(
        description=""
    )
    parser.add_argument(
        "--env-file", "-e",
        help=("File containing environment variables to be set before running "),
        default=".env",
    )
    parser.add_argument(
        "--loglevel", "-l",
        help=("Set the logging level"),
        default="INFO",
    )

    args = parser.parse_args()

    load_dotenv(args.env_file)

    numeric_level = getattr(logging, args.loglevel.upper())
    logging.basicConfig(level=numeric_level)
    logging.debug(f"Env file: {args.env_file}")

    logging.info("Transforming raw data...")
    transform(loglevel=numeric_level)
    logging.info("Transformation successful!")

if __name__ == "__main__":
    main()


import subprocess
import json
import os
import FAIRsoft
from FAIRsoft import utils

from os import listdir
from os.path import isfile, join


def run_repoenricher(base_path, out_path):
    # build repoenricher.pl command
    '''from 'pubmed-github-bitbucket/opeb-enrichers/repoEnricher/README.md'
    perl repoEnricher.pl --config myConfig.ini --directory=output
    '''
    config_path = '/'.join([base_path, 'config.ini'])
    repoEnricher='/'.join([base_path, 'repoEnricher.pl'])
    command_template = "perl {repoEnricher} --config {config} --directory={output}"
    command = command_template.format(repoEnricher=repoEnricher, config=config_path, output=out_path)
    print(command)

    # run command
    process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()

    return


def build_id(original_id):
    #parsing original id
    
    fields = original_id.split('/')

    if len(fields)>6:
        name_plus = fields[5]
        main_fields = name_plus.split(':')
        name = main_fields[1]
        if len(main_fields)>2:
            version =  main_fields[2]
        else:
            version = None
        
        type_ = fields[6]
        # building new id
        if version:
            name_version=':'.join([name, version])
        else:
            name_version = name
    
        template = 'https://openebench.bsc.es/monitor/tool/repository:{name_version}/{type_}'
        new_id = template.format(name_version=name_version, type_=type_)
        return(new_id)
    
    else:
        name = fields[5]
        template = 'https://openebench.bsc.es/monitor/tool/repository:{name}'
        new_id = template.format(name=name)
        return(new_id)
  

def import_data():
    # 1. run_enricher
    REPOENRICHER_PATH=os.getenv('REPOENRICHER_PATH', None)
    if REPOENRICHER_PATH:
        REPOENRICHER_OUTPUT_PATH=os.getenv('REPOENRICHER_OUTPUT_PATH', None)

    if REPOENRICHER_OUTPUT_PATH:
        run_repoenricher(REPOENRICHER_PATH, REPOENRICHER_OUTPUT_PATH)

    # 2. take all files in output folder, put "@source" in each and push to alambique
    OUTPUT_REPOS=os.getenv('OUTPUT_REPOS', 'repositories.json')

    repos_out_files = ['/'.join([OUTPUT_REPOS,f]) for f in listdir(OUTPUT_REPOS) if isfile(join(OUTPUT_REPOS, f))]
    if '/'.join([OUTPUT_REPOS,'manifest.json']) in repos_out_files:
        repos_out_files.remove('/'.join([OUTPUT_REPOS,'manifest.json']))

    # 3. Connection to database
    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')
    ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique')

    if STORAGE_MODE =='db':
        alambique = FAIRsoft.utils.connect_collection(ALAMBIQUE)
    else:
        OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
        OUTPUT_REPOS = os.getenv('OUTPUT_REPOS', 'opeb_metrics.json')
        output_file = OUTPUT_PATH + '/' + ALAMBIQUE + '/' + OUTPUT_REPOS

    log = {'names':[],
       'n_ok':0,
       'errors': []}
    # For file in output_path
    for out_file in repos_out_files:
        with open(out_file, 'r') as repo_entry:
            repo_json = json.load(repo_entry)
            # 4. add @source
            try:
                repo_json['@data_source'] = 'repository'
            except:
                print(repo_json)
            finally:
                # 5. insert in collection
                repo_json['@id'] = build_id(repo_json['@id'])
                #print(repo_json)
                if STORAGE_MODE=='db':
                    log = FAIRsoft.utils.push_entry(repo_json, alambique, log)
                else:
                    log = FAIRsoft.utils.save_entry(repo_json, output_file, log)
                    
if __name__ == '__main__':
    import_data()



import os
import FAIRsoft
class dMetadataFetcher():
    def __init__(self, tools_galaxy_metadata):
        self.repositories = tools_galaxy_metadata
        self.seen_tools = set()
        
    def get_dependencies(self, latest_revision):
        if latest_revision['tool_dependencies']:
            return(list(latest_revision['tool_dependencies'].keys()))
        else:
            return([])
    
    def retrieve_metadata(self, tool):
        if tool:
            latest_revision_id = max(iter(tool.keys()))
            latest_revision = tool[latest_revision_id]
            if 'tools' in latest_revision.keys():
                dependencies = self.get_dependencies(latest_revision)
                for t in latest_revision['tools']:
                    if latest_revision['tools'][0]['id']+latest_revision['tools'][0]['version'] not in self.seen_tools:
                        entry = {}
                        entry['id'] = latest_revision['tools'][0]['id']
                        entry['name'] = latest_revision['tools'][0]['name']
                        entry['version'] = latest_revision['tools'][0]['version']
                        entry['dependencies'] = dependencies
                        entry['@data_source'] = "galaxy_metadata"
                        self.seen_tools.add(latest_revision['tools'][0]['id']+latest_revision['tools'][0]['version'])
                        entry['@id'] = 'https://openebench.bsc.es/monitor/tool/galaxy_metadata:{name}:{version}/cmd'.format(name=entry['id'], version=entry['version'])
                        return(entry)
        return({})

    def process_metadata(self):
        STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')
        ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique')

        if STORAGE_MODE =='db':
            alambique = FAIRsoft.utils.connect_collection(ALAMBIQUE)
        else:
            OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
            OUTPUT_TOOLSHED=os.getenv('OUTPUT_TOOLSHED', 'toolshed.json')
            output_file = OUTPUT_PATH + '/' + ALAMBIQUE + '/' + OUTPUT_TOOLSHED

        log = {'names':[],
                'n_ok':0,
                'errors': []}

        for tool in self.repositories:
            entry = self.retrieve_metadata(tool)
            if entry:
                if STORAGE_MODE=='db':
                    log = FAIRsoft.utils.push_entry(entry, alambique, log)
                else:
                    log = FAIRsoft.utils.save_entry(entry, output_file, log)
            else:
                continue




import json
import pickle
import re
import logging
from munch import munchify

from FAIRsoft.utils import instance
from FAIRsoft.utils import setOfInstances
from FAIRsoft.utils import canonicalTool
from FAIRsoft.utils import canonicalSet
from FAIRsoft.utils import webTypes
from FAIRsoft.utils import update_entry, save_entry, push_entry

def save_obj(obj, name ):
    with open('obj/'+ name + '.pkl', 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def worker_pre_integration(i_n, inst, groupInst):
    name = inst.name
    if not inst.version:
        inst.version = 'unknown'
    if not inst.type:
        inst.type = 'unknown'

    if name in groupInst.keys():
        if inst.type in groupInst[name].keys():
            newList = groupInst[name][inst.type] + [inst]
            groupInst[name][ inst.type] = newList
        else:
            groupInst[name][inst.type]= [inst]  
    else:
        logging.debug(f"{inst.name} not in totalNames")
    
    return(groupInst)
    

def build_pre_integration_dict(setsOfInsts):
    '''
    input: setsOfInstances is a list of sets of instances
    by name and type
    '''
    totalNames = []
    names = []
    for instances in setsOfInsts:
        names.append(set([a.name for a in instances]))
        totalNames = totalNames + [ a.name for a in instances ]


    groupInst = dict()
    totalNames = set(totalNames)

    # Grouping the instances by name and type in a dictionary (groupInst)
    for name in totalNames: # iterating tool names 
        groupInst[name] = {}
    i_n = 0
    for set_ in setsOfInsts:
        for inst in set_:
            i_n+=1
            groupInst = worker_pre_integration(i_n, inst, groupInst)

    return(totalNames, groupInst)

# TODO
# Identify format
# Re-format
# Rmove duplicates
def identify_publication_format(pub):
    '''
    returns 'type'/'format' of publication information in entry
    possible formats types: 
        - 'oeb': typical OEB
        - 'key': { 'doi' : foo } or { 'pmid' : foo } or { 'pmid' : foo }
        - ''
    '''
    # OEB
    if type(pub) == dict:
        if 'entries' in pub:
            return('oeb')


def build_new_publication_meta(pubs_by_id):
    fields = ['doi','pmid','pmcid','url','citations','citation','title','year','journal','references','count']
    new_entry = {
            'doi':pubs_by_id.get('doi'),
            'pmid':pubs_by_id.get('pmid'),
            'pmcid':pubs_by_id.get('pmcid'),
            'url':None,
            'citations':None,
            'citation':None,
            'title':None,
            'year': None,
            'journal':None,
            'references':None,
            'count':None
        }
    for key in fields:
        key_found = False
        while key_found:
            for entry in pubs_by_id['entries']:
                if entry.get(key):
                    new_entry[key] = entry.get(key)
                    key_found = True

    return(new_entry)


def clean_reps_publications(publications):
    #unique_pubs_by_id = [{'ids': [doi, pmid, pmcid], 'entries':[publications]}]
    unique_pubs_by_id = []
    for pub_ in publications:
        if type(pub_) != list:
            pub_ = [pub_]
        for publication in pub_:
            #TODO extract ids if possible
            id_types = ['doi' ,'pmid', 'pmcid']
            # if doi/pmid/pmcid present

            publication = publication.__dict__

            ids_present = [id_type for id_type in id_types if id_type in publication.keys()]
            match = False
            while match:
                if ids_present:
                    for unique_entry in unique_pubs_by_id.keys():
                        for id_type in ids_present:
                            if publication[id_type] in unique_entry['ids']:
                                # publication already seen
                                if publication in unique_entry['entries']:
                                    # This publication DONE
                                    match = True
                                # publication not seen --> store
                                new_unique_by_id[unique_entry['entries']].append(publication)
                                # This publication DONE
                                match = True
                            
                    # If not in any unique_pubs_id[key]:
                    new_unique_by_id = {
                        'ids' : {
                            'doi':publication.get('doi'),
                            'pmid': publication.get('pmid'),
                            'pmcid': publication.get('pmcid')
                        },
                        'entries' : []
                    }
                    [publication.pop(id_type) for id_type in id_types]
                    new_unique_by_id.append(publication)

                    unique_pubs_by_id.append(new_unique_by_id)
                    # This publication DONE
                    match = True


                # if ids not present
                else:
                    new_unique_by_id = {
                        'ids' : {
                            'doi': None,
                            'pmid': None,
                            'pmcid': None
                        },
                        'entries' : [publication]
                    }
                    unique_pubs_by_id.append(new_unique_by_id)
                    match = True

        # Building publication instance def metadata:
        new_publications = []

        for pub in unique_pubs_by_id:
            new_publications.append(build_new_publication_meta(pub))

        return(new_publications)

def clean(item):
    if item == 'unknown':
        item = None
    if item == 'None':
        item = None
    
    return(item)

def clean_list(items_list):
    if None in items_list:
            items_list.remove(None)
    return(items_list)


def worker_integration(name, groupInst):
    new_insts = []
    for type_ in groupInst[name].keys():
       
        instaList =  groupInst[name][type_]
        version = [] 
        for a in instaList:
            if a.version:
                 version.append(clean(a.version[0]))
        version = list(set(version))
        
        newInst = instance(name, type_, version)

        newInst.label = [clean(a.label) for a in instaList if a.label]
        newInst.label = list(set(clean_list(newInst.label)))

        newInst.description = [clean(a.description[0]) for a in instaList if a.description] # constructing_consensus
        newInst.description = clean_list(newInst.description)

        newLinks = []
        for inst in instaList:
            for link in inst.links:
                if link not in newLinks and link:
                    newLinks.append(clean(link))
        newInst.links = clean_list(newLinks)

        
        pubs = []
        for pretool in instaList:
            if pretool.publication:
                for p in pretool.publication:
                    if p not in pubs:
                        pubs.append(clean(p))

        #newInst.publication =  clean_reps_publications(pubs)
        newInst.publication = clean_list(pubs)

        newInst.download = [clean(item) for sublist in instaList for item in sublist.download]  
        newInst.download = clean_list(newInst.download)

        inst_instr = [a.inst_instr for a in instaList]
        if True in inst_instr:
            newInst.inst_instr = True #
        else:
            newInst.inst_instr = False

        tests = [a.test for a in instaList]
        if True in tests:
            newInst.test = True #
        else:
            newInst.test = False

        newSrc = []
        for a in instaList:
            if a.src:
                for s in a.src:
                    if s!=None:
                        newSrc.append(clean(s))

        newInst.src = clean_list(newSrc)

        newOs = []
        for inst in instaList:
            if inst.os:
                for os in inst.os:
                    if os not in newOs:
                        newOs.append(clean(os))

        newInst.os = clean_list(newOs)


        newInputs = []
        for insta in instaList:
            if insta.input:
                for Dict in insta.input:
                    if Dict not in newInputs:
                        newInputs.append(clean(Dict))
                    else:
                        continue

        newInst.input = clean_list(newInputs) # list of strings/dicts
        # TODO: fix this inconsistency with input as strings or dicts in the generators

        newOutputs = []
        for insta in instaList:
            if insta.output:
                for Dict in insta.output:
                    if Dict not in newOutputs:
                        newOutputs.append(clean(Dict))
                    else:
                        continue
            # TODO: fix this inconsistency with output as strings or dicts in the generators

        newInst.output = clean_list(newOutputs) # list of strings

        newDep = []
        for inst in instaList:
            if inst.dependencies:
                for dep in inst.dependencies:
                    if dep not in newDep:
                        newDep.append(clean(dep))
        newInst.dependencies = clean_list(newDep) # list of strings

        newDocs = []
        for inst in instaList:
            if inst.documentation:
                for doc in inst.documentation:
                    if doc not in newDocs:
                        newDocs.append(clean(doc))
                    else:
                        continue

        newInst.documentation = clean_list(newDocs) # list of lists [[type, url], [type, rul], ...]

        newLicense = []
        for a in instaList:
            if a.license:
                for lic in a.license:
                    if lic:
                        if clean(lic) not in newLicense:
                            newLicense.append(clean(lic))
                else:
                    continue

        newInst.license = clean_list(newLicense)

        newInst.termsUse = False #
        newInst.contribPolicy = False

        newAuth  = []
        for l in instaList:
            if l.authors:
                for auth in l.authors:
                    try:
                        if clean(auth) and clean(auth.lstrip()) not in newAuth:
                            newAuth.append(clean(auth.lstrip()))
                        else:
                            continue
                    except Exception as e:
                        logging.error(e)
                        logging.error("Cannot clean authors:")
                        logging.error(l.authors)

        newInst.authors = clean_list(newAuth) # list of strings

        newRepos = []
        for t in instaList:
            if t.repository:
                for rep in t.repository:
                    if rep not in newRepos:
                        newRepos.append(clean(rep))
        newInst.repository = clean_list(newRepos)

        newSource = []
        for inst in instaList:
            if inst.source:
                for s in inst.source:
                    if s not in newSource:
                        newSource.append(clean(s))
        
        newInst.source = clean_list(newSource)
        
        newInst.operational = False
        newInst.ssl = False
        newInst.bioschemas = False
        for a in instaList:
            if a.operational == True:
                newInst.operational = True
            if a.ssl == True:
                newInst.ssl=True
            if a.bioschemas == True:
                newInst.bioschemas = True
        
        semantics = {'inputs':[],'outputs':[],'topics':[], 'operations':[]}
        for inst in instaList:
            if inst.semantics:
                for field in ['topics', 'operations', 'inputs', 'outputs']:
                    if inst.semantics[field]:
                        [semantics[field].append(item) for item in inst.semantics[field]]
        try:
            for field in ['topics', 'operations']:
                semantics[field] = list(set(semantics[field])) # remove duplic
        except Exception as e:
            logging.error("Error in merging semantics:")
            logging.error(f"Semantics: {semantics}")
            logging.error(f"Instances: {instaList}")

        newInst.semantics=semantics
        
        edam_topics = set()
        for inst in instaList:
            if inst.edam_topics:
                [edam_topics.add(item) for item in inst.edam_topics]
        
        newInst.edam_topics = list(edam_topics)

        edam_operations = set()
        for inst in instaList:
            if inst.edam_operations:
                [edam_operations.add(item) for item in inst.edam_operations]
        
        newInst.edam_operations = list(edam_operations)
        
        tags = set()
        for inst in instaList:
            if inst.tags:
                [tags.add(item) for item in inst.tags]
        newInst.tags = list(tags)

        # return instead of pushing to db
        inst=newInst.__dict__
        inst['@id'] = "https://openebench.bsc.es/monitor/tool/{name}:{version}/{type}".format(name=inst['name'], version=inst['version'], type=inst['type'])
        new_insts.append(inst)

    return(new_insts)


def integrate_types(groupOfInstances):
    types = list(groupOfInstances.keys())
    not_unknown_types = [type_ for type_ in  types if type_!='unknown']
    n_not_unknown = len(not_unknown_types)
    if n_not_unknown==1:
        known_type = not_unknown_types[0]
        new_known_insts = groupOfInstances[known_type] + groupOfInstances['unknown']
        groupOfInstances[known_type] = new_known_insts
        groupOfInstances.pop('unknown', None)
    return(groupOfInstances)


def create_integrated_instances(groupInst, out_resource, storage_mode='db'):
    # return resuting instances for testing purposes
    result = []
    n_i=0
    log = {
        'errors':[],
        'n_ok':0,
    }
    landmarks = {str(int((len(groupInst)/5)*i)): f"{i*10}%" for i in range(0,10)}
    # for each canonical tool name
    for name in groupInst.keys():
        n_i += 1
                
        if str(n_i) in landmarks.keys():
            logging.debug(f'{n_i}/{len(groupInst)} ({landmarks[str(n_i)]}) canonical tools processed\r')

        if 'unknown' in groupInst[name].keys():
            groupInst[name] = integrate_types(groupInst[name])
        
        integrated_insts = worker_integration(name, groupInst)
        
        # for each instance of the canonical tool, push to db or save to file
        for integrated_inst in integrated_insts:
            result.append(integrated_inst)

            if storage_mode=='db': 
                log = update_entry(integrated_inst, out_resource, log)
                #log = push_entry(integrated_inst, out_resource, log)
            elif storage_mode=='filesystem':
                log = save_entry(integrated_inst, out_resource, log)

    
    return(result, log)



def generateCanonicalSet(instsctDist):
    '''
    groups instances by name into canonicals
    input: {
             name1: [instance1, instance2],
             name2: [instance3],
             ...
             }
    output: canonicalSet
    '''
    newCanonSet  = canonicalSet()
    for name in instsctDist.keys():
        instances = instsctDist[name]
        sources = list(set([item for sublist in instances for item in sublist.source]))
        types = list(set([inst.type for inst in instances if inst.type != None ]))
        newCanon = canonicalTool(name, instances, sources, types)
        newCanonSet.addCanononical(newCanon)
    return(newCanonSet)



def loadJSON(path):
    with open(path) as fil:
        return(json.load(fil))

global stdFormats
def prepFAIRcomp(instances):
    stdFormats= getFormats(instances)
    return(stdFormats)
    
def get_stdFormats():
    return(stdFormats)


def getFormats(instances):
    inputs = [a.input for a in instances]
    inputs_ = [a for a in inputs]
    inputsNames = []

    nonSFormats = ['txt', 'text', 'csv', 'tsv', 'tabular', 'xml', 'json', 'nucleotide', 'pdf', 'interval' ]
    for List in inputs_:
        for eachD in List:
            if 'format' in eachD.keys():
                if ' format' not in eachD['format']['term'] and eachD['format']['term'].lstrip() not in nonSFormats:
                    if '(text)' not in eachD['format']['term']:
                        if eachD['format']['term'].lstrip() not in inputsNames:
                            inputsNames.append(eachD['format']['term'].lstrip())
    return(inputsNames)



import requests
import json

session = requests.Session()

class reposFetcher():

    def __init__(self):
        self.tools = []

    def fetch_tools(self):
        repositories_list = self._get_repositories_dict()
        for repository_dict in repositories_list:
            if repository_dict['type']=='unrestricted':
                self.tools.append(galaxyTool(repository_dict))
        self.all_metadatas = [tool.metadata for tool in self.tools if tool.metadata]

    def _get_repositories_dict(self):
        REPS_URL = "https://toolshed.g2.bx.psu.edu/api/repositories?" 
        repositories_json = get_url(REPS_URL)
        repositories_dict_list =  json.loads(repositories_json.text) 
        return(repositories_dict_list)
    
    def export_metadatas(self, output_file_name):
        all_metadatas = [tool.metadata for tool in self.tools if tool.metadata]
        with open(output_file_name, 'w') as outputfile:
            json.dump(all_metadatas, outputfile)

class galaxyTool():
    def __init__(self, repository_dict):
        self.metadata =  self._get_metadata(repository_dict)
        self.revisions = self.metadata.keys()

    def _get_metadata(self, repository_dict):
        url_string = "https://toolshed.g2.bx.psu.edu/api/repositories/{id_}/metadata?"
        id_ = repository_dict['id']
        url = url_string.format(id_ = id_)
        req = get_url(url)
        meta = req.json()
        return(meta)

def get_url(url):
    try:
        re = session.get(url)
    except:
        print('Impossible to make the request')
        print("problematic url: " + url)
        return(None)
    else:    
        return(re)


if __name__ == '__main__':
    RF = reposFetcher()
    RF.fetch_tools()
    RF.export_metadatas('galaxy_metadatas.json')


from functools import wraps
import os
import warnings
import time
import logging

# import pymongo/json only if necessary
if os.getenv('STORAGE_MODE', 'db') == 'db':
    from pymongo import MongoClient
else:
    import json

# --------------------------------------------
# Constants 
# --------------------------------------------
global webTypes
webTypes = ['rest', 'web', 'app', 'suite', 'workbench', 'db', 'soap', 'sparql']

# environment variables containing whether sources to be transformed in .env
sources_to_transform = [
    'BIOCONDUCTOR',
    'BIOCONDA',
    'BIOTOOLS',
    'TOOLSHED',
    'GALAXY_METADATA',
    'SOURCEFORGE',
    'GALAXY_EU',
    'OPEB_METRICS',
    'BIOCONDA_RECIPES',
    'BIOCONDA_CONDA',
    'REPOSITORIES'
]

# Labels of sources accross FAIRsoft pacakage. Must be consistent!!!! 
# Present in:
# 1.- 'sources' field in `instance`` objects (and anywhere they appear in code) 
# 2.- toolGenerators in FAIRsoft.meta_transformers
# if labels change in one place, they must change in the others for everythong to keep working

sources_labels = {
    'BIOCONDUCTOR':'bioconductor',
    'BIOCONDA':'bioconda',
    'BIOTOOLS':'biotools',
    'TOOLSHED':'toolshed',
    'GALAXY_METADATA':'galaxy_metadata',
    'SOURCEFORGE': 'sourceforge',
    'GALAXY_EU': 'galaxy',
    'OPEB_METRICS':'opeb_metrics',
    'BIOCONDA_RECIPES':'bioconda_recipes',
    'BIOCONDA_CONDA':'bioconda_conda',
    'REPOSITORIES': 'repository'
}

# --------------------------------------------
# Helper functions
# --------------------------------------------
def connect_collection(host:str='localhost', port:int=27017, db='test', collection:str='test'):
    '''Returns the collection object for the given collection name. 
    This is a helper function for any code that needs to access a collection.

    collection_name: name of collection to access
    Port, host and database are specified in .env file.    
    '''
    # localhost:27017 is the default port for mongodb

    client = MongoClient(host, port)
    collection = client[db][collection]

    return collection


def push_entry(tool:dict, collection:'pymongo.collection.Collection', log:dict):
    '''Push tool to collection.

    tool: dictionary. Must have at least an '@id' key.
    collection: collection where the tool will be pushed.
    log : {'errors':[], 'n_ok':0, 'n_err':0, 'n_total':len(insts)}
    '''
    # Push to collection
    # date objects cause trouble and are prescindable
    if 'about' in tool.keys():
            tool['about'].pop('date', None)
    try:
        insertResult = collection.insert_one(tool)
    except Exception as e:
        log['errors'].append({'file':tool,'error':e})
        logging.error(f"ERROR while pushing to database the following entry: {tool}. Exception: {e}")
    else:
        log['n_ok'] += 1
    finally:
        return(log)

def update_entry(tool:dict, collection:'pymongo.collection.Collection', log:dict):
    '''Update tool to collection.

    tool: dictionary. Must have at least an '@id' key.
    collection: collection where the tool will be pushed.
    log : {'errors':[], 'n_ok':0, 'n_err':0, 'n_total':len(insts)}
    '''
    # Push to collection
    # date objects cause trouble and are prescindable
    if 'about' in tool.keys():
            tool['about'].pop('date', None)
    try:
        updateResult = collection.update_one({'@id': tool['@id']}, {'$set': tool}, upsert=True)
    except Exception as e:
        log['errors'].append({'file':tool,'error':e})
        logging.error(f"ERROR while pushing to database the following entry: {tool}. Exception: {e}")

    else:
        log['n_ok'] += 1
    finally:
        return(log)


def save_many(tools, output_file, log):
    '''Save tool to file.

    tool: dictionary. Must have at least an '@id' key.
    output_file: file where the tool will be saved.
    log : {'errors':[], 'n_ok':0, 'n_err':0, 'n_total':len(insts)}
    '''
    # 4. Push to file
    # date objects cause trouble and are prescindable
    for tool in tools:
        if 'about' in tool.keys():
            tool['about'].pop('date', None)
    
    try:
        with open(output_file, 'w') as f:
            json.dump(tools, f)
    
    except Exception as e:
        log['errors'].append({'file':tool['name'],'error':e})
        raise
        # return(log)

    else:
        log['n_ok'] += 1
    finally:
        return(log)
        
        
def save_entry(tool, output_file, log):
    '''Save tool to file.

    tool: dictionary. Must have at least an '@id' key.
    output_file: file where the tool will be saved.
    log : {'errors':[], 'n_ok':0, 'n_err':0, 'n_total':len(insts)}
    '''
    # 4. Push to file
    # date objects cause trouble and are prescindable

    if 'about' in tool.keys():
            tool['about'].pop('date', None)
    try:
        if os.path.isfile(output_file) is False:
            with open(output_file, 'w') as f:
                json.dump([tool], f)
        else:
            with open(output_file, 'r+') as outfile:
                print('Saving to file: ' + output_file)
                data = json.load(outfile)
                data.append(tool)
                # Sets file's current position at offset.
                outfile.seek(0)
                json.dump(data, outfile)

    except Exception as e:
        log['errors'].append({'file':tool['name'],'error':e})
        raise
        # return(log)

    else:
        log['n_ok'] += 1
    finally:
        return(log)



def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')
        return result
    return timeit_wrapper


# --------------------------------------------
# Classes
# --------------------------------------------
class instance(object):

    def __init__(self, 
                 name='fairsoft_default_name', 
                 type_='fairsoft_default', 
                 version='fairsoft_default', 
                 dictionary=None):
        # initialize this class from a dictionary
        if dictionary != None:
            #print(dictionary)
            for key, value in dictionary.items():
                if key != '_id':
                    setattr(self,key,value)
        
        # initialize this class from name, type and version
        else:
            # Minimal initiallization of instance
            ## name is required. If not specified, a default value is assigned and a warning raised
            self.attribute_check_and_set('name', name)
            self.attribute_check_and_set('type', type_)

            ## version and type are strongly recommended. If not set, raise a warning and assign None.
            self.attribute_check_and_set('version', version)
            
            # remaining attributes are set to None/empty
            self.label : str = None
            self.links : List(str) = []
            self.publication : List() =  []
            self.download : List(str) = []
            self.inst_instr : bool = False
            self.test : bool = False
            self.src : List(str) = []
            self.os : List(str) = []
            self.input : List(dict) = [] #  {'format' : <format> , 'uri' : <uri> , 'data' : <data> , 'uri': <uri>}
            self.output : List(dict) = [] #  {'format' : <format> , 'uri' : <uri> }
            self.dependencies : List(str) = []
            self.documentation : List(list) = [] # [[type, url], [type, rul], ...]
            self.license : List(str) = []
            self.termsUse : List(str) = []
            self.contribPolicy = False
            self.authors : List(str) = []
            self.repository : List(str) = []
            self.description : List(str) = []
            self.source : List(str) = []
            self.bioschemas : bool  = False
            self.https : bool = False
            self.operational : bool = False
            self.ssl : bool = False
            self.edam_topics : List(str) = []
            self.edam_operations : List(str) = []
            self.semantics : dict = {}
            self.tags : List(str) = []
            

    def attribute_check_and_set(self, key, value):
        '''Check if the attribute is set.
        If it is not, set default attribute to the instance and raise a warning.
        '''
        if value == 'fairsoft_default':
            warnings.warn(f'Instance {key} not specified. Setting instance.{key} = None')
            setattr(self,key,None)

        elif value == 'fairsoft_default_name':
            # Default name instead of None for trackability/debugging purposes
            warnings.warn(f'Instance {key} not specified. Assigning default value "fairsoft_default_name"', Warning)
            setattr(self,key,value)

        else:
            setattr(self,key,value)


    def set_super_type(self):
        if self.type in webTypes:
            self.super_type='web'
        else:
            self.super_type='no_web'


    ##============ Findability metrics computation functions ================

    def compF1_2(self):
        '''
        Identifiability of version:
        Whether there is a scheme to uniquely and properly identify the software version.
        A version of the form X.X is considered acceptable: True. Anything else is False
        '''
        if self.version:
            vers_veredicts = []
            for v in self.version:
                if v:
                    if len(v.split('.'))==2:
                        vers_veredicts.append(True)
                    else:
                        vers_veredicts.append(False)
                # for now, a single True is enough, can be changed in the future, not allowing any False
            if True in vers_veredicts:
                return(True)
        return(False)

    global struct_meta
    # TODO this list must depend on the analyzed sources. Take from config.yaml
    struct_meta = ['biotools', 'bioconda', 'github', 'bitbucket', 'galaxy', 'toolshed', 'opeb_metrics']
    def compF2_1(self):
        '''
        Structured Metadata
        Metadata is adjusted to specific metdata formats
        The sources in struct_meta are structured. If these sources are among self.source: True. Otherwise: False
        '''
        if True in [a in struct_meta for a in self.source]:
            return(True)
        else:
            return(False)

    def compF2_2(self):
        '''
        Ontologies or controlled vocabularies used
        '''
        # look for EDAM terms
        if self.semantics:
            for k in self.semantics.keys():
                if self.semantics[k]:
                    return(True)
        return(False)


    global softReg
    softReg = ['biotools', 'bioconda', 'bioconductor']
    def compF3_1(self):
        '''
        Searchability in registries
        Whether software is included in the main software registries.
        If the source is among the software registries: True. Otherwise: Falsecource
        '''
        if True in [a in softReg for a in self.source]:
            return(True)
        else:
            return(False)

    def compF3_2(self):
        '''
        Searchabiliy in software repositories
        Whether software can be found in any of the major software repositories e.g. GitHub, GitLab, SourceForge, 
        If the instance has an associated repository uri: True. Otherwise: False
        '''
        if len(self.repository)>0:
            return(True)
        else:
            return(False)

    def compF3_3(self):
        '''
        Searchability in literature.
        Whether software can be found in specialized literatue services e.g. EuropePMC, PubMed, Journals Site, bioArxiv.
        If the instance at least one associated publication: True. Otherwise: False
        '''
        if len(self.publication)>0:
            return(True)
        else:
            return(False)

    ##============== Accessibility metrics computation functions ================

    def compA1_1(self):
        '''
        WEB
        Existance of API or web 
        Whether it is possible to access a working version of the tool through and API or web. 
        A 200 status when accessing the links provided is consired acceptable.
        '''
        return(self.operational)

    def compA1_2(self):
        '''
        NO WEB
        Existence of downloadable and buildable software working version
        If there is a download link: True ## we do not check if it is available. 
        '''
        if self.super_type == 'no_web':
            if len(self.download)>0:
                return(True)
            elif True in [s in self.source for s in ['bioconda', 'bioconductor','galaxy','toolshed','bioconda_conda','bioconda_recipes']]:
                return(True)
            else:
                return(False)
        else:
            return(False)

    def compA1_3(self):
        '''
        NO WEB
        Existence of installation instructions
        Whether there is a set of instructions and other necessary information the user can follow to build the software
        We check self.inst_instructions (already a boolean)
        '''
        #print('A1_3', end='\r')
        if self.super_type == 'no_web':
            return(self.inst_instr)
        else:
            return(False)

    def compA1_4(self):
        '''
        Existence of test data
        Whether test data is available
        We check self.test (already a boolean)
        '''
        return(self.test)

    def compA1_5(self):
        '''
        NO WEB
        Existence of software source code
        Whether software source code is available 
        '''
        if self.super_type == 'no_web':
            if len(self.src)>0:
                return(True)
            else:
                return(False)
        else:
            return(False)
    
    def compA3_1(self):
        '''
        WEB
        Registration not compulsory
        Whether homepage can be accessed without registration
        '''
        if self.super_type == 'web':
            return(self.operational)
        else:
            return(True)


    def compA3_2(self):
        '''
        NO WEB
        Availability of version for free OS
        Whether the software can be used in a free operative system
        '''
        if self.super_type == 'no_web':
            if 'Linux' in self.os:
                return(True)
            else:
                return(False)
        else:
            return(True)

    def compA3_3(self):
        '''
        No WEB
        Availability for several OS
        Whether there are versions of the software for several operative systems
        '''
        if self.super_type == 'no_web':
            if len(self.os)>1:
                return(True)
            else:
                return(False)
        else:
            return(False)

    def compA3_4(self):
        '''
        NO WEB
        Availability on free e-Infrastructures
        Whether the software can be used in a free e-infrastructure
        We are only considering galaxy servers and vre
        '''
        if self.super_type == 'no_web':
            eInfra = ['vre.multiscalegenomics.eu', 'galaxy.', 'usegalaxy.']
            for url in self.links:
                for e in eInfra:
                    if e in url:
                        return(True)
                else:
                    return(False)
            return(False)
        else:
            return(False)


    def compA3_5(self):
        '''
        NO WEB
        Availability on several e-Infrastructures
        Whether the software can be used in several e-infrastructure
        '''
        if self.super_type == 'no_web':
            count = 0
            eInfra = ['vre.multiscalegenomics.eu', 'galaxy.', 'usegalaxy.']
            for url in self.links:
                for e in eInfra:
                    if e in url:
                        count += 1
            if count>1:
                return(True)

            return(False)
        else:
            return(False)

    ##============== Interoperability metrics computation functions ================
    def compI1_1(self):
        '''
        Usage of standard data formats
        Whether the input and output datatypes are formally specified and related to accepted ontologies
        '''
        for i in self.input:
            if 'format' in i.keys():
                if i['format']['term'] in self.stdFormats:
                    return(True)

        for i in self.output:
            if 'format' in i.keys():
                if i['format']['term'] in self.stdFormats:
                    return(True)

        return(False)

    def compI1_3(self):
        '''
        Verificability of data formats
        Whether input/output data are specified using verifiable schemas (e.g. XDS, Json schema, ...)
        '''
        if self.compI1_1() == True:
            return(True)

        verifiable_formats = ['json', 'xml', 'rdf', 'xds']
        formats = self.input + self.output
        terms = set()
        for i in formats:
            if 'format' in i.keys():
                terms.add(i['format']['term'])
        
        for term in terms:
            if term in verifiable_formats:
                return(True)

        return(False)


        return(False)


    def compI1_4(self):
        '''
        Flexibility of data format supported
        Whether the software allows to choose among various input/output data formats, or provide the necessary tools to convert other common formats into the supported ones.
        '''
        ins = []
        formats = self.input + self.output
        for i in formats:
            if 'format' in i.keys():
                ins.append(i['format']['term'])

        if len(ins)>1:
            return(True)
        else:
            return(False)

    def compI2_1(self):
        '''
        Existence of API/library version 
        Whether the software has API /library versions to be included in users' pipelines
        '''
        interTypes = ['Library', 'Web API']
        for t in self.type:
            if t in interTypes:
                return(True)
        else:
            return(False)


    def compI2_2(self):
        '''
        E-infrastructure compatibility
        '''
        if 'galaxy' in self.source:
            return(True)
        elif 'toolshed' in self.source:
            return(True)
        else:
            return(False)

    def compI3_1(self):
        '''
        Dependencies statement
        Whether the software includes details about dependencies
        '''
        if len(self.dependencies)>0:
            return(True)
        else:
            return(False)

    def compI3_2(self):
        '''
        Dependencies are provided
        Whether the software includes its dependencies or mechanisms to access them
        '''
        # checking source
        if 'galaxyShed' in self.source:
            return(True)
        elif 'bioconda' in self.source:
            return(True)
        elif 'bioconductor' in self.source:
            return(True)
        
        #checking links
        sources_with_dependencies = ['bioconda', 'bioconductor', 'galaxy.']
        for url in self.links:
            if True in  [a in url for a in sources_with_dependencies]:
                return(True)
        
        return(False)

    # ===================== Reusability ==============================================================
    def compR1_1(self):
        '''
        Existence of usage guides
        Whether software user guides are provided
        '''
        noGuide = ['license', 'terms of use', 'news']
        for doc in self.documentation:
            for string in noGuide:
                if string not in doc[0].lower(): # doc[0] is the type of document
                    return(True)

        return(False)


    def compR2_1(self):
        '''
        Existence of license
        Whether license is stated
        '''
        
        for doc in self.documentation:
            if 'license' in doc[0].lower():
                print('In doc')
                return(True)
        
        if self.license:
            if len(self.license)>0:
                for a in self.license:
                    # some licenses are actually a list of one license in old versions of ETL
                    if type(a) == list:
                        a = a[0]
                    if a.lower()=='unlicensed' or a.lower()=='unknown' or a.lower()=='unlicense':
                        continue
                    else:
                        if a:
                            return(True)
        return(False)
        
    
    def compR2_2(self):
        '''
        Technical conditions of use
        '''
        for doc in self.documentation:
            if 'conditions of use' in doc[0].lower():
                return(True)
            elif 'terms of use' in doc[0].lower():
                return(True)
        if self.license:
            if len(self.license)>0:
                for a in self.license:
                    # some licenses are actually a list of one license in old versions of ETL
                    if type(a) == list:
                        a = a[0]
                    if a.lower()=='unlicensed' or a.lower()=='unknown' or a.lower()=='unlicense':
                        continue
                    else:
                        if a:
                            return(True)

        return(False)
        
    

    #def compR3_1(self):
    def compR3_2(self):
        '''
        Existence of credit
        Whether credit for contributions is provided
        '''
        if len(self.authors)>0:
            return(True)
        else:
            return(False)


    def compR4_1(self):
        '''
        Usage of version control
        Whether the software follows a version-control system
        '''
        for repo in self.repository:
            if 'github' in repo or 'mercurial-scm' in repo:
                return(True)
        return(False)



    def FAIRscores(self):
        self.scores = FAIRscores()
        # ===================== Findability =========================
        # F1
        self.scores.F1 = (0.8*self.metrics.F1_1 
                        + 0.2*self.metrics.F1_2)
        # F2
        self.scores.F2 = 0.6*self.metrics.F2_1 + 0.4*self.metrics.F2_2
        # F3
        acc = [self.metrics.F3_1, self.metrics.F3_2, self.metrics.F3_3].count(True)
        if acc == 1:
            self.scores.F3 = 0.7
        elif acc == 2:
            self.scores.F3 = 0.85
        elif acc == 3:
            self.scores.F3 = 1
        # F
        self.scores.F = (0.4*self.scores.F1
                       + 0.2*self.scores.F2
                       + 0.4*self.scores.F3)
        # ===================== Accessibility =======================
        if self.super_type == 'web':
            # A1
            self.scores.A1 = (0.6*self.metrics.A1_1 
                            + 0.4*self.metrics.A1_4)
            # A3
            self.scores.A3 = 1.0*self.metrics.A3_1 
        else:
            # A1
            self.scores.A1 = (0.5*self.metrics.A1_2 
                            + 0.2*self.metrics.A1_3 
                            + 0.1*self.metrics.A1_4 
                            + 0.2*self.metrics.A1_5)
            # A3
            self.scores.A3 = (1/4)*(self.metrics.A3_2
                                  + self.metrics.A3_3
                                  + self.metrics.A3_4
                                  + self.metrics.A3_5)
        # A
        self.scores.A = (0.7*self.scores.A1
                       + 0.3*self.scores.A3)
        #  A_2 not computable bc we do not have the appropriate metrics.
        #  self.A += (self.metrics.A2_1*(1/3)+self.metrics.A2_2*(2/3))*0.15
        
        # ===================== Interoperability =====================

        # I1
        self.scores.I1 = (0.5*self.metrics.I1_1
                        + 0.3*self.metrics.I1_2
                        + 0.3*self.metrics.I1_3
                        + 0.2*self.metrics.I1_4)
        # I2
        self.scores.I2 = (0.5*self.metrics.I2_1
                        + 0.5*self.metrics.I2_2)
        # I3
        self.scores.I3 = (1/3)*(self.metrics.I3_1
                              + self.metrics.I3_2
                              + self.metrics.I3_3)
        # I
        self.scores.I = (0.6*self.scores.I1
                       + 0.1*self.scores.I2
                       + 0.3*self.scores.I3)
        # ===================== Reusability ===========================

        # R1
        self.scores.R1 = 1.0*self.metrics.R1_1
        # R2
        if self.metrics.R2_1:
            self.scores.R2 = 1.0
        elif self.metrics.R2_2:
            self.scores.R2 = 1.0
        else:
            self.scores.R2 = 0.0
        # R3
        self.scores.R3 = 1.0*self.metrics.R3_2
        # R4
        self.scores.R4 = 1.0*self.metrics.R4_1

        #R
        self.scores.R = (0.3*self.scores.R1
                       + 0.3*self.scores.R2
                       + 0.2*self.scores.R3
                       + 0.2*self.scores.R4)

    def generateFAIRMetrics(self, stdFormats):
        self.stdFormats = stdFormats
        # FINDABILITY
        self.metrics = FAIRmetrics()

        self.metrics.F1_1 = True # all have a name
        self.metrics.F1_2 = self.compF1_2() #
        self.metrics.F2_1 = self.compF2_1()
        self.metrics.F2_2 = self.compF2_2() 

        self.metrics.F3_1 = self.compF3_1()
        self.metrics.F3_2 = self.compF3_2()
        self.metrics.F3_3 = self.compF3_3()
        # ACCESIBILITY
        self.metrics.A1_1 = self.compA1_1() # Existance of API or web 
        self.metrics.A1_2 = self.compA1_2() # Existance of downloadable and buildable software working version
        self.metrics.A1_3 = self.compA1_3() # Existance of installation instructions
        self.metrics.A1_4 = self.compA1_4() # Existance of test data
        self.metrics.A1_5 = self.compA1_5() # Existance of software source code

        self.metrics.A2_1 = False # Metadata of previous versions at software repositories
        self.metrics.A2_2 = False # Existence of accesible previous versions of the software

        self.metrics.A3_1 = self.compA3_1() # Registration compulsory
        self.metrics.A3_2 = self.compA3_2() # Availability of version for free OS
        self.metrics.A3_3 = self.compA3_3() # Availability for several OS
        self.metrics.A3_4 = self.compA3_4() # Availability on free e-Infrastructures
        self.metrics.A3_5 = self.compA3_4() # Availability on several e-Infrastructures

        self.metrics.I1_1 = self.compI1_1()  # Usage of standard data formats
        self.metrics.I1_2 = False # NOT FOR NOW # Usage of standard API framework
        self.metrics.I1_3 = self.compI1_3() # Verificability of data formats
        self.metrics.I1_4 = self.compI1_4() # Flexibility of data format supported
        self.metrics.I1_5 = False # NOT FOR NOW # Generation of provenance information

        self.metrics.I2_1 = self.compI2_1() # Existance of API/library version 
        self.metrics.I2_2 = self.compI2_2() # E-infrastructure compatibility

        self.metrics.I3_1 = self.compI3_1()
        self.metrics.I3_2 = self.compI3_2()
        self.metrics.I3_3 = self.compI3_2() # Same as befor, BY NOW

        self.metrics.R1_1 = self.compR1_1()
        self.metrics.R1_2 = False #NOT FOR NOW

        self.metrics.R2_1 = self.compR2_1()
        self.metrics.R2_2 = self.compR2_2()

        self.metrics.R3_1 = False # Not for now
        self.metrics.R3_2 = self.compR3_2()

        self.metrics.R4_1 = self.compR4_1()
        self.metrics.R4_2 = False # By now
        self.metrics.R4_3 = False # By now


class FAIRscores():
    def __init__(self):
        
        self.F = 0.0 
        self.F1 = 0.0
        self.F2 = 0.0
        self.F3 = 0.0

        self.A = 0.0
        self.A1 = 0.0
        self.A2 = 0.0
        self.A3 = 0.0

        self.I = 0.0
        self.I1 = 0.0
        self.I2 = 0.0
        self.I3 = 0.0

        self.R = 0.0
        self.R1 = 0.0
        self.R2= 0.0
        self.R3 = 0.0
        self.R4 = 0.0


class FAIRmetrics(object):

    def __init__(self):
        self.F1_1 = False # uniqueness of name
        self.F1_2 = False # idenfibiability of version

        self.F2_1 = False # structured metadata
        self.F2_2 = False # standarized metadata

        self.F3_1 = False # searchability in registries
        self.F3_2 = False # searchability in software repositories
        self.F3_3 = False # searchability in literature

        self.A1_1 = False
        self.A1_2 = False
        self.A1_3 = False
        self.A1_4 = False
        self.A1_5 = False

        self.A2_1 = False
        self.A2_2 = False

        self.A3_1 = False
        self.A3_2 = False
        self.A3_3 = False
        self.A3_4 = False
        self.A3_5 = False

        self.I1_1 = False
        self.I1_2 = False
        self.I1_3 = False
        self.I1_4 = False
        self.I1_5 = False

        self.I2_1 = False
        self.I2_2 = False

        self.I3_1 = False
        self.I3_2 = False
        self.I3_3 = False

        self.R1_1 = False
        self.R1_2 = False

        self.R2_1 = False
        self.R2_2 = False

        self.R3_1 = False
        self.R3_2 = False

        self.R4_1 = False
        self.R4_2 = False
        self.R4_3 = False

class canonicalSet(object):
    
    def __init__(self):
        self.canonicals = []

    def addCanononical(self, canon):
        self.canonicals.append(canon)

class canonicalTool(object):

    def __init__(self, name, instances, sources, types):
        self.name = name
        self.instances = instances
        self.sources = sources
        self.types = types

    def computeFAIRmetrics(self):
        self.F = max([ins.F for ins in self.instances])
        self.A = max([ins.A for ins in self.instances])
        self.I = max([ins.I for ins in self.instances])
        self.R = max([ins.R for ins in self.instances])

class setOfInstances(object):

    def __init__(self, source):
        self.source = source
        self.instances = []




import json
from pymongo import MongoClient


def add_source(entry):
    entry['@data_source'] = 'opeb_metrics' 
    return(entry)

def update_entries(entries):
    for i in range(len(entries)):
        entries[i] = add_source(entries[i])
    return(entries)

def push_to_DB(metrics):
    for entry in metrics:
        entry['@data_source'] = 'opeb_metrics' 
        updateResult = alambique.update({'@id':entry['@id']}, { '$set': entry }, upsert=True, multi=True)
    return

client = MongoClient(port=27017)
db = client.OPEB
alambique = db.alambique
path='/home/eva/FAIRsoft/FAIRsoft_ETL/tools_importers/data/metrics.json'

with open(path, 'r') as input: 
    try:
        metrics = json.load(input)
    except Exception as e:
        raise(e)
    else:
        push_to_DB(metrics)



import unittest
import json 
import os

import FAIRsoft
from FAIRsoft.integration.run_integration import collect_pretools_instances
from FAIRsoft.integration.integration import build_pre_integration_dict
from FAIRsoft.integration.integration import create_integrated_instances

# test  run_integration.py        
class TestRunIntegration(unittest.TestCase):

        def test_collect_pretools_instances(self):
            with open('data/pretools.json', 'r') as infile:
                pretools = json.load(infile)

            names = ['a4', 'a4base', 'a4classif', 'a4core', 'a4preproc', 'a4reporting', 'abaenrichment', 'abarray', 'absseq', 'abseqr']
            descriptions = []
            names = set(names)

            os.environ['BIOCONDUCTOR'] = 'True'
            insts = collect_pretools_instances(pretools, 'filesystem')
            calculated_names = []
            for source_list in insts:
                if source_list:
                    calculated_names.extend([inst.name for inst in source_list])
            
            #print(calculated_names)
            
            self.assertEqual(set(calculated_names), names)



# test integration.py
class TestIntegration(unittest.TestCase):
    
    def test_publications_integration(self):
        # input to test
        with open('data/pretools_integration.json', 'r') as infile:
            pretools = json.load(infile)
        # correct output (integrated tool)
        with open('data/tools_integration_assert.json', 'r') as infile:
            assert_tool = json.load(infile)
        
        STORAGE_MODE = 'filesystem'
        allInsts = collect_pretools_instances(pretools, STORAGE_MODE)
        
        totalNames, pre_integration_dict = build_pre_integration_dict(allInsts)
 

        ## Create integration intances. Returns a list of instances and log.
        resulting_tools, log = create_integrated_instances(pre_integration_dict, None, None)

        # --- For debugging purposes-----------------

        #print(json.dumps(resulting_tools[0], sort_keys=True, indent=4))
        #print('-----------------------------------------------------')
        #print(json.dumps(assert_tool[0], sort_keys=True, indent=4))

        #print('-----------------------------------------------------')


        for key,value in resulting_tools[0].items():
            if key not in assert_tool[0].keys():
                self.assertEqual(value, assert_tool[0][key])

        



import requests
import os
from pprint import pprint

token = os.getenv('GITHUB_TOKEN', '...')
owner = "MartinHeinz"
repo = "python-project-blueprint"
query_url = f"https://api.github.com/repos/{owner}/{repo}/issues"
params = {
    "state": "open",
}
headers = {'Authorization': f'token {token}'}
r = requests.get(query_url, headers=headers, params=params)
pprint(r.json())


import requests as re
import json
import configparser
import os
import pymongo
from pymongo import MongoClient

import FAIRsoft
from FAIRsoft import utils

def getURLcontent(url):
    try:
        res = re.get(url)

    except:
        print("Data from OPEB download failed")
        return(None)

    else:
        return(res)

def add_source(entry):
    entry['@data_source'] = 'opeb_metrics' 
    return(entry)

def update_entries(entries):
    for i in range(len(entries)):
        entries[i] = add_source(entries[i])

    return(entries) 

def save_JSON(response, out_filename):
    # 2 steps: decoding and writing
    # we could use only one json.dump()
    with open(out_filename, 'w') as outfile:
        try:
            content_decoded=json.loads(response.text) 
        except:
            print("Could not decode opeb db JSON")
        else:
            entries = update_entries(content_decoded)
            json.dump(entries, outfile)

def push_to_DB(response, collection):
    # 2 steps: decoding and writing
    # we could use only one json.dump()
    try:
        content_decoded=json.loads(response.text) 
    except:
        print("Could not decode opeb db JSON")
    else:
        entries = update_entries(content_decoded)
        collection.insert_many(entries)
    return

def parse_config(configfile):
    config = configparser.ConfigParser()
    config.read(configfile)
    params = {
        'metrics_url': config['URLS']['METRICS'],
        'output_dir' : config['PATHS']['OUTDIR']
    }
    return(params)

def import_data():
    # 1. connect to DB/get output files
    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')

    if STORAGE_MODE =='db':
        ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique') 
        client = MongoClient('localhost', 27017)
        alambique = client['observatory2'][ALAMBIQUE]

    else:
        OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
        OUTPUT_OPEB_TOOLS = os.getenv('OUTPUT_OPEB_TOOLS', 'opeb_tools.json') 
        output_file = OUTPUT_PATH + '/' + ALAMBIQUE + '/' + OUTPUT_OPEB_TOOLS


    # 2. Get metrics metadata from OPEB
    URL_OPEB_METRICS = os.getenv('URL_OPEB_METRICS', 'https://openebench.bsc.es/monitor/metrics/')
    json_res = getURLcontent(URL_OPEB_METRICS)
    
    # Decode response
    try:
        content_decoded=json.loads(json_res.text)
    except:
        raise Exception(f'Could not decode opeb metrics JSON. Please, check URL: {URL_OPEB_METRICS}')
    else:
        entries = update_entries(content_decoded)
        log = {'names':[],
           'n_ok':0,
           'errors': []}
        # output tools metadata
        for inst_dict in entries:
            if STORAGE_MODE=='db':
                print('pushing to db')
                try:
                    updateResult = alambique.update_many({'@id':inst_dict['@id']}, { '$set': inst_dict }, upsert=True)
                except Exception as e:
                    log['errors'].append({'file':inst_dict,'error':e})
                    return(log)
                else:
                    log['n_ok'] += 1
            else:
                log = FAIRsoft.utils.save_entry(inst_dict, output_file, log)

if __name__ == "__main__":
    import_data()


from modulefinder import STORE_GLOBAL
import sys
import time
import pickle
import os
import logging

import json
from pymongo import MongoClient
from munch import munchify

from FAIRsoft.integration.integration import build_pre_integration_dict
from FAIRsoft.integration.integration import create_integrated_instances

from FAIRsoft.utils import timeit
from FAIRsoft.utils import instance
from FAIRsoft.utils import sources_labels
from FAIRsoft.utils import sources_to_transform
from FAIRsoft.utils import connect_collection


def open_json_collection(file_name: str):
    """Opens a JSON file
    
    file_name: path to the JSON file
    Used to open <pretools.json>  and <tools.json>
    """
    with open(file_name, 'r') as f:
        data = json.load(f)
        return(data)


def collect_pretools_instances(pretools, STORAGE_MODE):
    """Collects all the instances in pretools from the sources specified in configuration (.env)

    pretools: either a 'pymongo.collection.Collection' or a list of dictionaries
    """

    sources = sources_to_transform

    allInsts = []
    N_instances = 0

    for source in sources:
        if os.getenv(source, 'True') == 'True':
            logging.info("Collecting instances from %s"%(source))
            this_source_label = sources_labels[source]

            # Dependign on the storage_mode (db/file), get instances apply appropriate method to get data
            if STORAGE_MODE=='db':
                insts_iter = pretools.find({'source': this_source_label})
            else:
                insts_iter = [tool_dict for tool_dict in pretools if tool_dict['source'] == [this_source_label]]

            # turn to instance object
            insts = [instance(dictionary = inst) for inst in insts_iter]
            
            allInsts.append(insts)
            names=set([inst.name for inst in insts])
            N_instances += len(insts)
            
            logging.info(f"Entries from {this_source_label}: {len(insts)}")
            logging.info(f"Generic tools from {this_source_label}: {len(names)}")

    logging.info(f"Total instances:{N_instances}")    
    return(allInsts)

def open_pickle(filename):
    with open(filename, 'rb') as f:
        x = pickle.load(f)
        return(x)

def save_to_json(item, filename):
    with open(filename, 'w') as fp:
        json.dump(item, fp)

def save_names_to_file(totalNames, filename):
    with open(filename, 'w') as outfile:
        for name in totalNames:
            outfile.write("%s\n"%(name))

def save_insts_by_name(inst_name_dict, filename):
    """Saves a dictionary with key=tool name and value=list(instances of that tool)
    
    inst_name_dict is a dictionary of instances by tool name. 
    # {name_1: [instance1, instance2, ...], [instance1'], [instance1'', instance2'', ...], ...}
    """
    with open(filename, 'w') as fp:
        json.dump(inst_name_dict, fp)
    
@timeit
def run_integration(loglevel=logging.INFO):
    logging.basicConfig(level=loglevel)

    HOST = os.getenv('HOST', 'localhost')
    PORT = os.getenv('PORT', 27017)
    DB = os.getenv('DB', 'test')
    PRETOOLS = os.getenv('PRETOOLS', 'pretools') 
    TOOLS = os.getenv('TOOLS', 'tools') 

    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')
    if STORAGE_MODE == 'db':
        # un-integrated collection of instances (input instances)
        pretools = connect_collection(host=HOST, port=int(PORT), db=DB, collection=PRETOOLS)
        # integrated collection of instances (output instances)
        tools = connect_collection(host=HOST, port=int(PORT), db=DB, collection=TOOLS)
    else:
        pretools_file = os.getenv('OUTPUT_PATH', './') + '/' + PRETOOLS + '.json'
        pretools = open_json_collection(pretools_file)
        tools_file = os.getenv('OUTPUT_PATH', './') + '/' + TOOLS + '.json'
        tools = open_json_collection(tools_file)


    ######----------- Data restructuring and integration ------------------------------------##########

    # 1. Collecting instances from pretools
  
    allInsts = collect_pretools_instances(pretools, STORAGE_MODE)
    
    totalNames, pre_integration_dict = build_pre_integration_dict(allInsts)

    ## Saving pre_integration_dict and names. 
    ## This step is not necessary, but it is useful for debugging
    # save_to_json(pre_integration_dict.copy(), 'pre_integration_dict.json')
    # save_names_to_file(totalNames, 'totalNames.txt')


    # 2. Integration of "tool" instances of same ID (name, type)
    n = []
    for k in pre_integration_dict.keys():
        for kk in pre_integration_dict[k].keys():
            n.append(len(pre_integration_dict[k][kk]))

    # 3. Create integrated instances and push to db/output to file.
    result, log = create_integrated_instances(pre_integration_dict, tools, STORAGE_MODE)
 

if __name__ == '__main__':
    run_integration()
    


import requests
import os
from bs4 import BeautifulSoup

import FAIRsoft
from FAIRsoft import utils

def get_entries(soup, projects):
    results = soup.find_all('div', attrs={"class":"result-heading-texts"}) # results panel
    #entries = results.find('a')
    for entry in results:
        e = entry.find('a')
        projects.append("https://sourceforge.net" + e['href'])
    return(projects)


def get_next(soup):
    URL=os.getenv('URL_SOURCEFORGE_PACKAGES', 'https://sourceforge.net/directory/science-engineering/bioinformatics/')
    try:
        next_href = soup.find('li', attrs={"class":"pagination-next"}).find('a')['href']
        page = next_href.split('/')[-1]
        next_url = f'{URL}{page}'
        return(next_url)
    except:
        return(None)        

def get_url(url):
    session = requests.Session()
    try:
        re = session.get(url)
    except:
        print('Impossible to make the request')
        print("problematic url: " + url)
        return(None)
    else:    
        return(re)
    
def get_soup(url):
    re = get_url(url)
    if re:
        soup = BeautifulSoup(re.text, 'lxml')
        return(soup)
    else:
        return(None)

def get_lastUpdate(project_soup):
    last_update = project_soup.find('time', attrs={"class":"dateUpdated"})['datetime']
    return(last_update)

def get_description(project_soup):
    description = project_soup.find('p', attrs={"itemprop":"description", "class":"description"})
    #print(description)
    if description != None:
        description_plain = description.text
    else:
        description_plain = None
    return(description_plain)

def get_homepage(project_soup):
    a = project_soup.find('a', attrs={"id":"homepage"})
    if a:
        homep = a['href']
    else:
        homep = None
    #print(homep)
    return(homep)
    
def get_project_info(project_soup):
    project_info = {}
    info = project_soup.find_all('section', attrs={"class":"project-info"})
    licens = []
    project_info['license'] = []
    project_info['registered'] = False
    project_info['categories'] = []

    for section in info:
        if section.header:
            #print(section.header.h4)
            if section.header.h4.text == 'Registered':
                registered = section.section.text.strip()
                project_info['registered'] = registered

        elif section.h3:
            if section.h3.text == "License":
                for a in section.find_all("a"): 
                    licens.append(a.text)
                project_info['license'] = licens
            if section.h3.text == "Categories":
                for a in section.find_all("a"):
                    project_info['categories'].append(a.span.text)
                print(f"Categories: {project_info['categories']}")
    
    return(project_info)

def get_OS(project_soup):
    OS = []
    platforms = project_soup.find_all('div', attrs={"class":"platforms"})
    for span in platforms:
        if span.meta:
            os = span.text.strip()
            OS.append(os)
    if len(OS)>0 and OS[0]!='':
        oss = OS[0]
        opsys = oss.split('\n')
        return(opsys)


def import_data(db_config, config):
    # 1. connect to DB/ set files
    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')
    ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique')

    if STORAGE_MODE =='db':
        alambique = FAIRsoft.utils.connect_collection(ALAMBIQUE)
    else:
        OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
        OUTPUT_SOURCEFORGE=os.getenv('OUTPUT_SOURCEFORGE', 'sourceforge.json')
        output_file = OUTPUT_PATH + '/' + ALAMBIQUE + '/' + OUTPUT_SOURCEFORGE

    
    # Go through pages and get all entries
    session = requests.HTMLSession()
    url=os.getenv('URL_SOURCEFORGE_PACKAGES', 'https://sourceforge.net/directory/science-engineering/bioinformatics/')
    
    projects = []
    while url:
        re = session.get(url)
        soup = BeautifulSoup(re.text, 'lxml')
        projects = get_entries(soup, projects)
        url = get_next(soup)
    
    print(f"Number of bioinformatics linux projects in SourceForge{len(projects)}")

    log = {'names':[],
       'n_ok':0,
       'errors': []}
    # Extract information from each entry
    for entry in projects:
        name = entry.split('/')[-2]
        entry_all = {}
        soup = get_soup(entry)
        if soup:
            entry_all['last_update'] = get_lastUpdate(soup)
            entry_all['description'] = get_description(soup)
            info = get_project_info(soup)
            entry_all['registered'] = info["registered"]
            entry_all['license'] = info["license"]
            entry_all['operating_systems'] = get_OS(soup)
            entry_all['repository'] = 'https://sourceforge.net/projects/' + name
            entry_all['homepage'] = get_homepage(soup)
            entry_all['name'] = name
            
            entry_all['@id'] = 'https://openebench.bsc.es/monitor/tool/sourceforge:{name}'.format(name=name)
            entry_all['@data_source'] = 'sourceforge'
            entry_all['@source_url'] = 'https://sourceforge.net/projects/' + name

            if STORAGE_MODE=='db':
                log = FAIRsoft.utils.push_entry(entry_all, alambique, log)
            else:
                log = FAIRsoft.utils.save_entry(entry_all, output_file, log)


if __name__ == "__main__":

    import_data()



from bs4 import BeautifulSoup
from bs4.dammit import EncodingDetector
import requests

session = requests.Session()
headers = {"User-Agent":"Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit 537.36 (KHTML, like Gecko) Chrome",
"Accept":"text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8"}

def getHTML(url, verb=False):
    '''
    Takes and url as an input and returns the corresponding
    bs4 object
    '''
    
    try:
        re = session.get(url, headers=headers, timeout=(10, 30))
    except:
        print(r'Cannot access HTML')
        return(None)
    else:
        if re.status_code == 200:
            # dealing with encoding
            http_encoding = re.encoding if 'charset' in re.headers.get('content-type', '').lower() else None
            html_encoding = EncodingDetector.find_declared_encoding(re.content, is_html=True)
            encoding = html_encoding or http_encoding

            # generating BeautifulSoup object
            bsObj = BeautifulSoup(re.content, 'html5lib', from_encoding=encoding)

            if verb == True:
                print("The title of html is %s"%bsObj.title.getText())
            return(bsObj)
        else:
            return(None)

def get_whole_table():
    table_url = "https://bioconductor.org/packages/release/bioc/"
    packages_lst_html = getHTML(table_url) # bs4 object

    table = packages_lst_html.find("div", {"class":"do_not_rebase"})
    if len(table)>1:
        print("Warning: more than one table found")
        
    return(table)


def extract_name_url(table):
    bioc_url= 'https://bioconductor.org/packages/release/bioc/'
    all_packages = []
    for tag in table.tbody:
        a = tag.find('a')
        if a and a!=-1:
            #name = a.get_text()
            url = bioc_url + a['href']
            print(url)
            all_packages.append(url)
    
    return(all_packages)


def fetch_all_packages():
    table = get_whole_table()
    all_packages = extract_name_url(table)
    return(all_packages)

if __name__ == "__main__":
    packs = fetch_all_packages()
    with open("tmp/list_packages.txt", "w") as outfile:
        print("writing URLs to list")
        x = 1
        for p in packs:
            print("Package %d of %d"%(x, len(packs)), end='\r')
            outfile.write("%s\t%s\n"%(p, 'NA'))
            x += 1



import json
from pymongo import MongoClient

# connecting to DB
client = MongoClient(port=27017)
db = client.OPEB
alambique = db.alambique

with open("xml_metadatas.json", "r") as metadatasf:
    metadatas = json.load(metadatasf)

for main_tool in metadatas:
    for tool in main_tool:
        tool["data_source"] = "toolshed"
        alambique.insert_one(tool)



import os
import importlib
import logging 

import FAIRsoft
import FAIRsoft.utils
from FAIRsoft.transformation.meta_transformers import tool_generators


def get_raw_data_db(source, alambique):
    logging.debug(f"Extracting raw data of source {source} from collection {alambique.name}")
    raw = alambique.find({"@data_source":source})
    return list(raw)

   
def output_bioconda_dict(generator, source):
    if source == 'bioconda_recipes':
        generator.type_dictionary('bioconda_types.json')


def open_raw_files(source):
    import json
    files_sources = {
        'BIOCONDUCTOR':'OUTPUT_BIOCONDUCTOR',
        'BIOCONDA':'OUTPUT_OPEB_TOOLS',
        'BIOTOOLS':'OUTPUT_OPEB_TOOLS',
        'TOOLSHED':'OUTPUT_TOOLSHED',
        'GALAXY_METADATA':'OUTPUT_TOOLSHED',
        'SOURCEFORGE':'OUTPUT_SOURCEFORGE',
        'GALAXY_EU': 'OUTPUT_OPEB_TOOLS',
        'OPEB_METRICS': 'OUTPUT_OPEB_METRICS',
        'BIOCONDA_RECIPES':'OUTPUT_BIOCONDA_RECIPES',
        'BIOCONDA_CONDA':'OUTPUT_BIOCONDA_CONDA',
        'REPOSITORIES': 'OUTPUT_REPOS',
    }
    try:
        OUTPUT_PATH = os.getenv('OUTPUT_PATH', './data')
        data_file = OUTPUT_PATH    + '/' + os.getenv(files_sources[source])
    except KeyError:
        print("No file for {}".format(source))
        raise
    except TypeError:
        print("No file for {}".format(files_sources[source]))
        raise
    else:
        with open(data_file, 'r') as infile:
            raw = json.load(infile)

        logging.info(f"Extracted raw data from {data_file}")
        logging.info(f"{len(raw)} entries from {files_sources[source]}")

        return raw


def add_to_pretools_file(insts, output_file):
    import json
    if not os.path.isfile(output_file):
            with open(output_file, 'w') as outfile:
                json.dump(insts, outfile)
    else:   
        with open(output_file) as json_data_file:
            pretools = json.load(json_data_file)
        pretools.append(insts)


def transform_this_source(raw, this_source_label):
    # Instantiate toolGenerator specific to this source
    logging.debug(f"Instantiating toolGenerator for {this_source_label}")
    generator_module = importlib.import_module(f".meta_transformers", 'FAIRsoft.transformation')
    
    logging.debug(f"Transforming raw data into instances")
    generator = generator_module.tool_generators[this_source_label](raw)
    
    # Export types dictionary in case of bioconda raw data 
    output_bioconda_dict(generator, this_source_label)
    
    # From instance objects to dictionaries
    insts = [i.__dict__ for i in generator.instSet.instances]
    logging.debug(f"Transformed {len(insts)} instances")
    return(insts)


def transform(loglevel, **kwargs):

    logging.basicConfig(level=loglevel)
    logging.getLogger('bibtexparser').setLevel(logging.WARNING)

    # Set environment variables:
    HOST = os.getenv('HOST', 'localhost')
    PORT = os.getenv('PORT', 27017)
    DB = os.getenv('DB', 'test')
    ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique')
    PRETOOLS = os.getenv('PRETOOLS', 'pretools') 

    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')
    if STORAGE_MODE =='db':
        logging.info('Storage mode: `database`')
        # raw data (input) collection
        logging.debug(f'Connecting to raw data collection {ALAMBIQUE} in database {DB}. Host: {HOST}:{PORT}')
        alambique = FAIRsoft.utils.connect_collection(host=HOST, port=int(PORT), db=DB, collection=ALAMBIQUE)
        # transformed data collection
        logging.debug(f'Connecting to raw data collection {PRETOOLS} in database {DB}. Host: {HOST}:{PORT}')
        pretools = FAIRsoft.utils.connect_collection(host=HOST, port=int(PORT), db=DB, collection=PRETOOLS)

    else:
        # Path of output file where the transformed data will be stored
        logging.info('Storage mode: `file`')
        logging.debug(f'Output file: {PRETOOLS}.json')
        output_file = os.getenv('OUTPUT_PATH', './') + '/' + PRETOOLS + '.json'

    # Run whole transformation pipeline
    for source in FAIRsoft.utils.sources_to_transform:
        logging.debug('Next source: {}'.format(source))

        # Check if source has to be transformed
        if os.getenv(source) == 'True':
            # label to match "source" field in the raw data and appropriate transformer
            this_source_label = FAIRsoft.utils.sources_labels[source]
                        
            # 1. getting raw data
            logging.info(f"------------- {source} -------------")
            logging.debug(f"Accessing data from `{source}` in database")
            if STORAGE_MODE =='db':
                raw_data = get_raw_data_db(this_source_label, alambique)
            else:
                raw_data = open_raw_files(source)
            
            if not raw_data:
                logging.info(f"No entries found for {this_source_label}")
            else:
                # 2. transformation
                logging.info(f"Transforming raw tools metadata from {this_source_label}")
                insts = transform_this_source(raw_data, this_source_label)
                
                # 3. output transformed data
                if STORAGE_MODE =='db':
                    log = {'errors':[], 'n_ok':0, 'n_err':0, 'n_total':len(insts)}
                    logging.debug("Updating database")
                    n=0
                    landmarks = {str(int((len(insts)/5)*i)): f"{i*10}%" for i in range(0,10)}
                    for inst in insts:
                        if str(n) in landmarks.keys():
                            logging.debug(f'{n}/{len(insts)} ({landmarks[str(n)]}) instances pushed to database\r')

                        inst['@id'] = f"https://openebench.bsc.es/monitor/tool/{inst['source'][0]}:{inst['name']}:{inst['version'][0]}/{inst['type']}"
                        log = FAIRsoft.utils.update_entry(inst, pretools, log)
                        #log = FAIRsoft.utils.push_entry(inst, pretools, log)
                        n+=1
                    
                    logging.info(f"Pushed {log['n_ok']} entries to database")
           
                else:
                    add_to_pretools_file(insts, output_file)    
                    logging.info(f"Saved to file")
        else:
            logging.info(f"Skipping {source}")
            

if __name__ == "__main__":
    
    transform()


import unittest
import json 
import warnings


import FAIRsoft
import FAIRsoft.transformation.meta_transformers as meta_trans 

class TestMetaTransformers(unittest.TestCase):
    
    def test_biocondaRecipesGen(self):
        """
        Bioconda recipes to instances
        """
        with open('data/bioconda_recipes.json', 'r') as datafile:
            data = json.load(datafile)
        
        biocondaGenerator = meta_trans.biocondaRecipesToolsGenerator(data)
        instances = biocondaGenerator.instSet.instances 

        # Correct number of instances and types
        assert_count_instances = {
            'thesias':{
                'insts':1, 
                'type':set(['cmd'])
                },
            'cmappy':{
                'insts': 2,
                'type':set(['lib', 'cmd'])
                },
            'ag.db' : {
                'insts':1,
                'type':set(['lib'])
            },
            'tbl2asn':{
                'insts':1, 
                'type':set(['cmd'])
            }
        }

        #  To test: Number of instances and types in generated instances
        test_count_instances = {}
        for inst in instances:
            if inst.name not in test_count_instances:
                test_count_instances[inst.name] = {
                    'insts': 1,
                    'type': set([inst.type])
                }
            else:
                test_count_instances[inst.name]['insts'] += 1
                test_count_instances[inst.name]['type'].add(inst.type)
        
        for  inst_name in test_count_instances:
            self.assertEqual(test_count_instances[inst_name], assert_count_instances[inst_name])

    def test_repository(self):
        '''
        Repository
        '''
        with open('data/bioconda_recipes.json', 'r') as datafile:
            data = json.load(datafile)

        assert_repos = {
            'thesias' : ["https://github.com/daissi/thesias"],
            'cmappy' : ["https://github.com/cmap/cmapPy"],
            'ag.db' : [],
            'tbl2asn': []
        }
        
        biocondaGenerator = meta_trans.biocondaRecipesToolsGenerator(data)
        instances = biocondaGenerator.instSet.instances

        for inst in instances:
            self.assertEqual(inst.repository, assert_repos[inst.name])

    def test_type_dict(self):
        with open('data/bioconda_recipes.json', 'r') as datafile:
            data = json.load(datafile)

        assert_dict = {
            'thesias': ['cmd'],
            'cmappy': ['lib', 'cmd'], 
            'ag.db': ['lib'],
            'tbl2asn': ['cmd']
            }

        biocondaGenerator = meta_trans.biocondaRecipesToolsGenerator(data)
        self.assertEqual(biocondaGenerator.type_dictionary(None), assert_dict)
    
    def test_OPEB_conda_types(self):
        '''
        Types from OPEB, OPEB metrics and conda that are also 
        present in bioconda recipes are assigned the recipe's type 
        in transformation steps
        '''
        with open('data/bioconda_transform.json', 'r') as datafile:
            data_al = json.load(datafile)

        assert_dict = {
            "rbcbook1": ["lib"],
            "fastq-filter": ["lib", "cmd"],
            "gargammel-slim": ["cmd"]
            }

        entries = {}

        for source in meta_trans.tool_generators:
            entries[source] = []

        for entry in data_al:
            if entry.get('@data_source') != None:
                for s in entries:
                    if entry['@data_source'] == s:
                        entries[s].append(entry)
        
        for source in entries:
            # the following sources must have the same type
            if source in ['conda', 'opeb_metrics', 'bioconda', 'bioconda_conda', 'bioconda_recipes']:
                if entries[source]:
                    generator = meta_trans.tool_generators[source](entries[source])
                    for inst in generator.instSet.instances:
                        self.assertIn(inst.type, assert_dict[inst.name])
            else:
                if entries[source]:
                    generator = meta_trans.tool_generators[source](entries[source])
                    for inst in generator.instSet.instances:
                        if inst.type not in assert_dict[inst.name]:
                            msg = f'\nType of "{inst.name}" according to:\n - bioconda recipe: {assert_dict[inst.name]}\n - biotools: {inst.type}\n'
                            warnings.warn(msg)
        
    def test_repostIntegrationType(self):
        '''
        Metadata from repositories is assigned the correct type
        bioconda - according to bioconda recipes
        galaxy - if original is 'workflow', then replace with cmd.
        else - keep the type as in the @id
        '''
        with open('data/repository_transform.json', 'r') as datafile:
            data = json.load(datafile)

        assert_dict = {
            "https://openebench.bsc.es/monitor/tool/repository:bridge-galaxy/workbench" : 'workbench',
            "https://openebench.bsc.es/monitor/tool/repository:bridge-galaxy/workflow" : 'cmd',
            "https://openebench.bsc.es/monitor/tool/repository:bridge-galaxy": None
        }

        reposGenerator = meta_trans.repositoryToolsGenerator(data)
        instances = reposGenerator.instSet.instances

        for entry in data:
            reposGenerator = meta_trans.repositoryToolsGenerator([entry])
            instances = reposGenerator.instSet.instances
            id_ = entry['@id']
            if instances:
                self.assertEqual(instances[0].type, assert_dict[id_])
            else:
                self.assertEqual(None, assert_dict[id_])

    def test_publications_obeb_metrics(self):
        '''
        opeb metrics publications must be transformed to a list ob objects
        [
            {
                cit_count: ...,
                doi: ...,
                pmcid: ...,
                year: ...,
                citations: ...,
                ...
            }
        ]
        '''
        with open('data/opeb_pubs_transform.json', 'r') as datafile:
            data_al = json.load(datafile)
        
        opeb_metrics_Generator = meta_trans.metricsOPEBToolsGenerator(data_al)
        instances = opeb_metrics_Generator.instSet.instances
        inst = instances[0]
        publications = inst.publication
        for pub in publications:
            #print(pub.keys())
            self.assertEqual(type(pub), dict)




if __name__ == '__main__':
    unittest.main()


import re
import os

url_id_bioc = {}
links = set()

def clean_url(url):
    URL_BIOCONDUCTOR=os.getenv('URL_BIOCONDUCTOR', 'https://bioconductor.org/packages/.*/bioc/.*')
    if re.search(URL_BIOCONDUCTOR, url):
        print(url)


def get_all_ids_links(tools):
    for tool in tools:
        if 'bioconductor' in tool['homepage']:
            url_id_bioc[tool['homepage']] = tool['@id']
    return(url_id_bioc)






import json
import os
import FAIRsoft
from FAIRsoft import utils

from .repos_metadata_importer import reposFetcher
from .galaxy_metadata import dMetadataFetcher
from .repos_config_importer import toolShedMetadataFetcher

def import_data(only_new):   

    #1. Fetch galaxy metadata
    print('Fetching Galaxy Toolshed Repositories Metadata...')
    RF = reposFetcher()
    RF.fetch_tools()
    repositories_metadata = RF.all_metadatas

    OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
    ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique')
    GALAXY_METADATA=os.getenv('GALAXY_METADATA', 'galaxy_metadatas.json')
    metadata_out=OUTPUT_PATH + '/' + ALAMBIQUE + '/' + GALAXY_METADATA

    RF.export_metadatas(metadata_out)
    
    #2. Parse and push fetched metadata
    print('Parsing Galaxy Toolshed Repositories Metadata...')
    with open(metadata_out, 'r') as meta_file:
        repositories_metadata = json.load(meta_file)
    
    dMFetcher = dMetadataFetcher(repositories_metadata)
    dMFetcher.process_metadata()
    
    #4. Download and process configuration files in repos
    print('Fetching Galaxy Toolshed Metadata inside repositories...')
    toolShedMDF = toolShedMetadataFetcher(tools_galaxy_metadata = repositories_metadata, 
                                          only_new=only_new)
    toolShedMDF.get_toolShed_files()


if __name__ == '__main__':
    import_data(False)


import argparse
import json
from matplotlib.font_manager import findfont
import requests
import ssl
import os
import pymongo
from pymongo import MongoClient
import FAIRsoft

from dotenv import load_dotenv

load_dotenv()


session = requests.Session()
headers = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit 537.36 (KHTML, like Gecko) Chrome",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
}

def getHTML(url, verb=False):
    ssl._create_default_https_context = ssl._create_unverified_context
    try:
        req = session.get(url, headers=headers, timeout=(20, 50), verify=False)
        # req = urllib.request.urlopen(url)
    except Exception as e:
        print(e)
        print(e)
        return None
    else:
        return req

def get_source(id_):
    string = id_.split('/')[5]
    source = string.split(':')[0]
    return(source)

def get_bioconda_biotools_galaxy_tools(tool, log):
    if tool['@id'].count('/')>5:
        source = get_source(tool['@id'])
        if source == 'biotools':
            tool['@data_source'] = 'biotools'
        elif source == 'bioconda':
            tool['@data_source'] = 'bioconda'
        elif source == 'galaxy':
            tool['@data_source'] = 'galaxy'

        tool['@source_url'] = tool['@id']
    else:
        log['canonical_N'] +=1

    return(tool, log)


def import_data():
    # 0. connect database/set output files
    STORAGE_MODE = os.getenv('STORAGE_MODE', 'db')

    if STORAGE_MODE =='db':
        ALAMBIQUE = os.getenv('ALAMBIQUE', 'alambique') 
        client = MongoClient('localhost', 27017)
        alambique = client['observatory2'][ALAMBIQUE]

    else:
        OUTPUT_PATH = os.getenv('OUTPUT_PATH', './')
        OUTPUT_OPEB_TOOLS = os.getenv('OUTPUT_OPEB_TOOLS', 'opeb_tools.json') 
        output_file = OUTPUT_PATH + '/' + ALAMBIQUE + '/' + OUTPUT_OPEB_TOOLS


    # 1. Download all opeb
    URL_OPEB_TOOLS = os.getenv('OPEB_URL', 'https://openebench.bsc.es/monitor/tool')
    print(f'OpenEBench tools URL: {URL_OPEB_TOOLS}')
    re = getHTML(URL_OPEB_TOOLS)

    # 2. Get tools
    tools = re.json()
    log = {'errors':[], 'n_ok':0, 'names': [],'canonical_N': 0}
    #For tool in OPEB Tool db
    print('tools obtained')
    for tool in tools:
        # 3. Process metadata
        tool, log = get_bioconda_biotools_galaxy_tools(tool,log)
        # 4. push to db/file
        if STORAGE_MODE=='db':
            print('pushing to db')
            if 'about' in tool.keys():
                tool['about'].pop('date', None)
            try:
                updateResult = alambique.update_many({'@id':tool['@id']}, { '$set': tool }, upsert=True)
            except Exception as e:
                log['errors'].append({'file':tool,'error':e})
                return(log)
            else:
                log['n_ok'] += 1


        else:
            log = FAIRsoft.utils.save_entry(tool, output_file, log)
     
    print(log)

    # Importation finished
    print(f'''\n----- OPEB Tools Importation finished -----
    Number of tools in OPEB {len(log['names'])}
    Number of canonical tools: {log['canonical_N']}''')
    print('Exceptions\n')
    for e in log['errors']:
        print(e['error'])


if __name__ == '__main__':
    import_data()


#!/usr/bin/env python3

from distutils.core import setup

setup(name='FAIRsoft',
      version='0.1.4',
      description='Library for the aggregation of Life Sciences software metadata and FAIR evaluation.',
      description_content_type='text/markdown',
      log_description='''FAIRsoft indicators are a set of FAIRness indicators for research software, specifically devised to be assesed automatically. 
      This package allows the user to gather metadata about research software specific to Life Sciences, harmonize and integrate it and then perform an evaluation of their compliance with FAIRsoft indicators.''',
      author='Eva Martin del Pico',
      author_email='eva.martin@bsc.es',
      license='AGLP-3.0',
      classifiers=['Development Status :: 4 - Beta'],
      install_requires = [
        'bidict',
        'matplotlib',
        'munch',
        'pymongo',
        'requests',
        'selenium',
        'simplejson',
        'webdriver_manager'],
      url='https://gitlab.bsc.es/inb/elixir/software-observatory/FAIRsoft_ETL/-/tree/master/FAIRsoft',
      packages=['FAIRsoft', 
                'FAIRsoft.importers',
                'FAIRsoft.importers.bioconductor_imp',
                  'FAIRsoft.importers.bioconda_imp',
                  'FAIRsoft.importers.bioconda_biotools_galaxy_imp',
                  'FAIRsoft.importers.toolshed_imp',
                  'FAIRsoft.importers.repositories_imp',
                  'FAIRsoft.importers.opeb_metrics_imp',
                  'FAIRsoft.importers.sourceforge_imp',
                'FAIRsoft.integration',
                'FAIRsoft.transformation',
                'FAIRsoft.indicators_evaluation'],
     )
